/* device independent bitmap */
#include <sp/spBaseLib.h>
#include <sp/spPluginP.h>
#include <sp/spCanvasP.h>
#include <sp/spGraphicsP.h>
#include <sp/spDrawP.h>

#include <sp/spDIBitmapP.h>

static spDIBitmapFuncList sp_default_dibitmap_func_list =
{
    NULL,
    NULL,
	
    spInitDIBitmapInstanceArch,
    spFreeDIBitmapInstanceArch,
	
    spGetDIBitmapPriorityArch,
    spGetDIBitmapFileTypeListArch,
    spGetDIBitmapFileDescListArch,
    spGetDIBitmapFileFilterListArch,
    spIsDIBitmapSupportedArch,
    spSetDIBitmapFileTypeArch,
    spGetDIBitmapFileTypeArch,
    NULL,
    NULL,
    NULL,
    spGetDIBitmapBestSuffixArch,
	
    spGetDIBitmapCapsArch,
	
    spOpenDIBitmapFileArch,
    spCloseDIBitmapFileArch,
	
    NULL,
    NULL,
    NULL,
    spGetDIBitmapInfoArch,
	
    spReadDIBitmapArch,
    spWriteDIBitmapArch,
	
    NULL,
    NULL,
};
    
void spDumpDIBitmapValue(spDIBitmap dibitmap)
{
    int x, y;
    int r, g, b;
    spBool locked;
    spDIBitmapInfoInternal *info;

    if (dibitmap == NULL) return;

    locked = dibitmap->info.locked;

    if (locked == SP_FALSE) spLockDIBitmap(dibitmap);

    info = (spDIBitmapInfoInternal *)&dibitmap->info;
    r = g = b = 0;
    
    for (y = 0; y < info->height; y++) {
	printf("%d: ", y);
	for (x = 0; x < info->width; x++) {
	    spGetDIBitmapRGB(dibitmap, x, y, &r, &g, &b);
	    /*printf("#%02x%02x%02x ", r, g, b);*/
	    printf("(%d,%d,%d) ", r, g, b);
	}
	printf("\n");
    }
    printf("\n");
    
    if (locked == SP_FALSE) spUnlockDIBitmap(dibitmap);
    
    return;
}

spDIBitmapFuncList *spGetDefaultDIBitmapFuncList(void)
{
    return &sp_default_dibitmap_func_list;
}

typedef struct _spDIBitmapHandlerList *spDIBitmapHandlerList;
struct _spDIBitmapHandlerList
{
    spDIBitmapFuncList *func_list;
    
    char *name;
    int index;
    int priority;
    unsigned long caps;

    int file_type_index;
    char *file_type;
    char *file_desc;
    char *file_filter;

    spDIBitmapHandlerList prev;
    spDIBitmapHandlerList next;
};

static spDIBitmapHandlerList allocDIBitmapHandlerList(const char *name, int index, spDIBitmapHandlerList prev_list)
{
    spDIBitmapHandlerList list;

    list = xalloc(1, struct _spDIBitmapHandlerList);
    list->name = strclone(name);
    list->index = index;
    list->priority = SP_PLUGIN_PRIORITY_NULL;
    list->caps = 0L;

    list->file_type_index = 0;
    list->file_type = NULL;
    list->file_desc = NULL;
    list->file_filter = NULL;

    list->prev = NULL;
    if (prev_list != NULL) {
	prev_list->next = list;
	list->prev = prev_list;
    }
    list->next = NULL;

    return list;
}

static void freeDIBitmapHandlerList(spDIBitmapHandlerList list)
{
    if (list == NULL) return;

    if (list->next != NULL) freeDIBitmapHandlerList(list->next);

    if (list->name != NULL) {
	xfree(list->name);
    }
    if (list->file_type != NULL) {
	xfree(list->file_type);
    }
    if (list->file_desc != NULL) {
	xfree(list->file_desc);
    }
    if (list->file_filter != NULL) {
	xfree(list->file_filter);
    }

    return;
}

static spDIBitmapHandlerList sp_dibitmap_handler_list = NULL;

static spDIBitmapHandlerList createDIBitmapHandlerList(spDIBitmapFuncList *func_list, const char *name, int index,
                                                       int priority, unsigned long caps, int file_type_index, 
                                                       const char **file_type_list, const char **file_desc_list, const char **file_filter_list,
                                                       spDIBitmapHandlerList prev_list)
{
    spDIBitmapHandlerList list;
    
    spDebug(100, "createDIBitmapHandlerList", "in\n");
    
    list = allocDIBitmapHandlerList(name, index, prev_list);
    list->func_list = func_list;
    list->priority = priority;
    list->caps = caps;
    list->file_type_index = file_type_index;
    if (file_type_list != NULL) {
	list->file_type = strclone(file_type_list[file_type_index]);
    } else {
	list->file_type = NULL;
    }
    if (file_desc_list != NULL) {
	list->file_desc = strclone(file_desc_list[file_type_index]);
    } else {
	list->file_desc = NULL;
    }
    if (file_filter_list != NULL) {
	list->file_filter = strclone(file_filter_list[file_type_index]);
    } else {
	list->file_filter = NULL;
    }

    if (sp_dibitmap_handler_list == NULL) {
	sp_dibitmap_handler_list = list;
	spAddExitCallback((spExitCallbackFunc)freeDIBitmapHandlerList, sp_dibitmap_handler_list);
    }

    spDebug(100, "createDIBitmapHandlerList", "done\n");
    
    return list;
}

static spBool createAllDIBitmapHandlerList(void)
{
    int i, j, k;
    int handler_count;
    const char *handler_name;
    void *instance;
    int priority;
    unsigned long caps;
    const char **file_type_list;
    const char **file_desc_list;
    const char **file_filter_list;
    spDIBitmapFuncList *func_list;
    spDIBitmapHandlerList list, prev_list;

    func_list = spGetDefaultDIBitmapFuncList();
    prev_list = NULL;

    for (i = 0;; i++) {
	if (func_list->get_count == NULL) {
	    handler_count = 1;
	} else {
	    handler_count = func_list->get_count();
	}
	spDebug(80, "createAllDIBitmapHandlerList", "i = %d, handler_count = %d\n", i, handler_count);

	for (j = 0; j < handler_count; j++) {
	    if (func_list->get_name == NULL) {
		handler_name = "default";
	    } else {
		handler_name = func_list->get_name(j);
	    }
	    spDebug(80, "createAllDIBitmapHandlerList", "j = %d, handler_name = %s\n", j, handler_name);
	    
	    if ((instance = func_list->init_instance(handler_name)) != NULL) {
		if (func_list->get_priority != NULL) {
		    priority = func_list->get_priority(instance);
		} else {
		    priority = SP_PLUGIN_PRIORITY_LOWEST;
		}
		spDebug(100, "createAllDIBitmapHandlerList", "priority = %d\n", priority);
		
		if ((file_type_list = func_list->get_file_type_list(instance)) != NULL) {
		    for (k = 0; file_type_list[k] != NULL; k++) {
			caps = 0L;
			if (func_list->set_file_type == NULL
			    || func_list->set_file_type(instance, k) == SP_TRUE) {
			    if (func_list->get_caps != NULL) {
				caps = func_list->get_caps(instance);
			    }
			}
			spDebug(100, "createAllDIBitmapHandlerList", "caps = %lx, file_type_list[%d] = %s\n",
				caps, k, file_type_list[k]);

			if (func_list->get_file_desc_list != NULL) {
			    file_desc_list = func_list->get_file_desc_list(instance);
			} else {
			    file_desc_list = NULL;
			}
			if (func_list->get_file_filter_list != NULL) {
			    file_filter_list = func_list->get_file_filter_list(instance);
			} else {
			    file_filter_list = NULL;
			}

			list = createDIBitmapHandlerList(func_list, handler_name, j, priority, caps, k, 
                                                         file_type_list, file_desc_list, file_filter_list, prev_list);

			prev_list = list;
		    }
		}
		func_list->free_instance(instance);
	    } else {
		spDebug(100, "createAllDIBitmapHandlerList", "init_instance failed\n");
	    }
	}

	if (i == 0) {
	    func_list = spGetPluginDIBitmapFuncList();
	} else {
	    break;
	}
    }

    spDebug(100, "createAllDIBitmapHandlerList", "done\n");
    
    return SP_TRUE;
}

static spDIBitmapHandler allocDIBitmapHandler(spDIBitmapFuncList *func_list, const char *name,
                                              unsigned long caps, int file_type_index, void *instance)
{
    spDIBitmapHandler handler;
    
    handler = xalloc(1, struct _spDIBitmapHandler);
    
    handler->func_list = func_list;
    handler->name = strclone(name);
    handler->caps = caps;
    handler->file_type_index = file_type_index;
    handler->instance = instance;
    memset(&handler->file_info, 0, sizeof(spDIBitmapInfo));

    return handler;
}

static spBool freeDIBitmapHandler(spDIBitmapHandler handler)
{
    if (handler->instance != NULL && handler->func_list->free_instance != NULL) {
	handler->func_list->free_instance(handler->instance);
    }
    if (handler->name != NULL) {
	xfree(handler->name);
    }
    xfree(handler);
    
    return SP_TRUE;
}

unsigned long spCalcDIBitmapRequiredCaps(spDIBitmapInfoInternal *info, unsigned long strict_mask)
{
    unsigned long caps = 0L;

    if (info == NULL) return 0L;

    if ((strict_mask & SP_DI_BITMAP_STRICT_YUV_STATE)
	&& (info->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK)) {
	caps |= SP_DI_BITMAP_CAPS_YUV;
    }
    if ((strict_mask & SP_DI_BITMAP_STRICT_GRAYSCALE_STATE)
	&& (info->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK)) {
	caps |= SP_DI_BITMAP_CAPS_GRAYSCALE;
    }
    if ((strict_mask & SP_DI_BITMAP_STRICT_SINGLE_RGB_STATE)
	&& (info->pixel_format & SP_DI_PIXEL_FORMAT_SINGLE_RGB_MASK)) {
	caps |= SP_DI_BITMAP_CAPS_SINGLE_RGB;
    }
    if ((strict_mask & SP_DI_BITMAP_STRICT_ALPHA_STATE)
	&& (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)) {
	caps |= SP_DI_BITMAP_CAPS_ALPHA;
	if (!(info->pixel_format & SP_DI_PIXEL_FORMAT_RGB_MASK)) {
	    caps |= SP_DI_BITMAP_CAPS_ALPHA_ONLY;
	}
    }
    if (strict_mask & SP_DI_BITMAP_STRICT_ALPHA_ORDER) {
	if (info->pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK) {
	    caps |= SP_DI_BITMAP_CAPS_FIRST_ALPHA;
	} else if (info->pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK) {
	    caps |= SP_DI_BITMAP_CAPS_LAST_ALPHA;
	}
    }
    if ((strict_mask & SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE)
	&& (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK)) {
	caps |= SP_DI_BITMAP_CAPS_ALPHA_PREMULTIPLIED;
    }
    if (strict_mask & SP_DI_BITMAP_STRICT_FLOAT_STATE
	&& (info->pixel_format & SP_DI_PIXEL_FORMAT_FLOAT_MASK)) {
	caps |= SP_DI_BITMAP_CAPS_FLOAT;
    }
    if (strict_mask & SP_DI_BITMAP_STRICT_PIXEL_ORDER) {
	if (info->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK) {
	    caps |= SP_DI_BITMAP_CAPS_INVERSE_FORMAT;
	} else {
	    caps |= SP_DI_BITMAP_CAPS_NONINVERSE_FORMAT;
	}
    }
    if (strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE) {
	if (info->pixel_stride == 3) {
	    caps |= SP_DI_BITMAP_CAPS_PIXEL_STRIDE_3;
	} else if (info->pixel_stride == 4) {
	    caps |= SP_DI_BITMAP_CAPS_PIXEL_STRIDE_4;
	}
    }
    if (strict_mask & SP_DI_BITMAP_STRICT_BIT_SIZE) {
	if (info->bit_size == 32) {
	    caps |= SP_DI_BITMAP_CAPS_BIT_SIZE_32;
	} else if (info->pixel_stride == 16) {
	    caps |= SP_DI_BITMAP_CAPS_BIT_SIZE_16;
	}
    }
    if (strict_mask & SP_DI_BITMAP_STRICT_UPSIDE_DOWN_STATE) {
	if (info->upside_down == SP_FALSE) {
	    caps |= SP_DI_BITMAP_CAPS_UPSIDE_UP;
	} else {
	    caps |= SP_DI_BITMAP_CAPS_UPSIDE_DOWN;
	}
    }

    return caps;
}

spBool spIsDIBitmapSupportedByHandler(spDIBitmapHandler handler, unsigned long caps, unsigned long options, const char *filename, const char *mode)
{
    const char *filename2 = NULL;

    if (handler == NULL) return SP_FALSE;
    
    spDebug(100, "spIsDIBitmapSupportedByHandler", "caps = %lx, handler->caps = %lx, options = %lx\n", caps, handler->caps, options);
    
    if (mode != NULL && mode[0] == 'w') {
	if (filename == NULL) {
	    return SP_FALSE;
	}
	filename2 = spGetSuffix(filename);
	caps |= SP_DI_BITMAP_CAPS_WRITE;
    } else if (mode != NULL && mode[0] == 'r') {
	if (filename == NULL) {
	    return SP_FALSE;
	}
	caps |= SP_DI_BITMAP_CAPS_READ;
    } else {
	filename2 = filename;
    }
    
    if ((handler->caps & caps) == caps) {
	if (filename == NULL) {
	    return SP_TRUE;
	} else {
	    spDebug(100, "spIsDIBitmapSupportedByHandler", "filename2 = %s\n", filename2);

	    return handler->func_list->is_supported(handler->instance, filename2);
	}
    }

    return SP_FALSE;
}

spDIBitmapHandler spGetDIBitmapSupportedHandler(unsigned long caps, unsigned long options,
                                                const char *filename, const char *mode, int *current_priority)
{
    int i, j;
    int priority;
    void *instance;
    const char *filename2 = NULL;
    spDIBitmapHandler handler = NULL;
    spBool write_mode = SP_FALSE;
    spBool supported = SP_FALSE;
    spDIBitmapHandlerList list = NULL;

    if (sp_dibitmap_handler_list == NULL) {
	createAllDIBitmapHandlerList();
    }

    if (filename != NULL) {
	if (mode != NULL && mode[0] == 'w') {
	    filename2 = spGetSuffix(filename);
	    write_mode = SP_TRUE;
	    caps |= SP_DI_BITMAP_CAPS_WRITE;
	} else {
	    filename2 = filename;
	    caps |= SP_DI_BITMAP_CAPS_READ;
	}
	spDebug(100, "spGetDIBitmapSupportedHandler", "filename = %s, filename2 = %s\n", filename, filename2);
    } else if (mode != NULL && mode[0] == 'w') {
        write_mode = SP_TRUE;
        caps |= SP_DI_BITMAP_CAPS_WRITE;
    }

    if (current_priority != NULL) {
	priority = MAX(*current_priority, 0);
    } else {
	priority = 0;
    }

    for (j = priority; j <= SP_PLUGIN_PRIORITY_LOWEST; j++) {
	spDebug(100, "spGetDIBitmapSupportedHandler", "j = %d\n", j);
	if (handler != NULL) {
	    break;
	}
	list = sp_dibitmap_handler_list;
	handler = NULL;
	
	for (i = 0; list != NULL; i++) {
	    spDebug(100, "spGetDIBitmapSupportedHandler", "i = %d, j = %d, list->priority = %d\n",
		    i, j, list->priority);

	    if (options & SP_DI_BITMAP_OPTION_AVOID_NATIVE_BASED_IO) {
		if (write_mode) {
		    if (list->caps & SP_DI_BITMAP_CAPS_WRITE_USE_BITMAP_ARGUMENT) {
			continue;
		    }
		} else {
		    if (list->caps & (SP_DI_BITMAP_CAPS_READ_USE_BITMAP_ARGUMENT | SP_DI_BITMAP_CAPS_READ_REPLACE_BITMAP)) {
			continue;
		    }
		}
	    }
	    
	    if (j == list->priority) {
		if ((instance = list->func_list->init_instance(list->name)) != NULL) {
		    supported = SP_FALSE;
		    spDebug(100, "spGetDIBitmapSupportedHandler", "list->file_type_index = %d, list->caps = %lx, options = %lx\n",
			    list->file_type_index, list->caps, options);

		    if (list->func_list->set_file_type != NULL) {
			list->func_list->set_file_type(instance, list->file_type_index);
		    }

		    if ((list->caps & caps) == caps) {
			if (filename2 == NULL) {
			    supported = SP_TRUE;
			} else {
			    supported = list->func_list->is_supported(instance, filename2);
			}
		    }

		    spDebug(100, "spGetDIBitmapSupportedHandler", "supported = %d\n", supported);
		    
		    if (supported) {
			handler = allocDIBitmapHandler(list->func_list, list->name,
                                                       list->caps, list->file_type_index, instance);
			break;
		    } else {
			list->func_list->free_instance(instance);
		    }
		} else {
		    spDebug(100, "spGetDIBitmapSupportedHandler", "init_instance failed\n");
		}
	    }

	    list = list->next;
	}
    }
    
    if (current_priority != NULL) {
	*current_priority = j;
    }

    spDebug(100, "spGetDIBitmapSupportedHandler", "done: handler = %ld\n", (long)handler);
    
    return handler;
}

spDIBitmap spCreateDIBitmapCore(spDIBitmapInfo *info, spDIBitmapHandler handler, unsigned long options)
{
    spDIBitmap dibitmap;

    dibitmap = xalloc(1, struct _spDIBitmap);
    memset(dibitmap, 0, sizeof(struct _spDIBitmap));
    if (info != NULL) {
	memcpy(&dibitmap->info, info, sizeof(spDIBitmapInfo));
    } else {
	dibitmap->info.pixel_format = SP_DI_PIXEL_FORMAT_DEFAULT;
	dibitmap->info.bit_size = 8;
	if (dibitmap->info.pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) {
	    dibitmap->info.pixel_stride = 4;
	} else {
	    dibitmap->info.pixel_stride = 3;
	}
    }
    dibitmap->handler = handler;

    if (options & SP_DI_BITMAP_STRICT_PIXEL_STRIDE) {
	if ((dibitmap->info.pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK)
	    || (dibitmap->info.pixel_format & SP_DI_PIXEL_FORMAT_LAST_ALPHA_MASK)) {
	    dibitmap->info.pixel_stride = 4;
	} else {
	    dibitmap->info.pixel_stride = 3;
	}
    }

    return dibitmap;
}

unsigned char *spAllocDIBitmapMemoryNonnative(spDIBitmapInfoInternal *info)
{
    long buf_size;
    unsigned char *buf;

    buf_size = info->line_stride * info->height;
    spDebug(80, "spAllocDIBitmapMemoryNonnative", "pixel_stride = %ld, line_stride = %ld, height = %ld, buf_size = %ld\n",
            info->pixel_stride, info->line_stride, info->height, buf_size);
    
    if ((buf = xalloc(buf_size, unsigned char)) != NULL) {
	info->pixel_format |= SP_DI_BITMAP_TYPE_NONNATIVE_MASK;
	return buf;
    } else {
	return NULL;
    }
}

spDIBitmap spCreateDIBitmapFromFileWithOptions(const char *filename, spDIBitmapInfo *info, unsigned long options)
{
    spDIBitmap dibitmap;
    unsigned long caps;
    unsigned long strict_mask;
    unsigned long strict_mask2;
    spDIBitmapHandler handler;
    spDIBitmapFuncList *func_list;
    int x, y;
    int r, g, b, a;
    spBool host_upside_down;
    spBool updown_swap;
    spBool use_nonnative = SP_FALSE;
    spBool flag = SP_FALSE;

    if (spIsTopLevelCreated() == SP_FALSE) {
	spDebug(10, "spCreateDIBitmapFromFileWithOptions", "toplevel has not been created\n");
	return NULL;
    }
    
    spDebug(80, "spCreateDIBitmapFromFileWithOptions", "options = %ld, filename = %s\n",
	    options, filename);

    if ((handler = spGetDIBitmapSupportedHandler(0L, options, filename, "r", NULL)) == NULL) {
	spDebug(10, "spCreateDIBitmapFromFileWithOptions", "spGetDIBitmapSupportedHandler failed\n");
	return NULL;
    }
    func_list = handler->func_list;

    dibitmap = spCreateDIBitmapCore(info, handler, options);

    caps = dibitmap->handler->caps;
    spDebug(80, "spCreateDIBitmapFromFileWithOptions", "caps = %lx\n", caps);

    if (func_list->open != NULL && func_list->get_info != NULL && func_list->read != NULL
	&& func_list->open(dibitmap->handler->instance, filename, "r") == SP_TRUE) {
	if (func_list->get_file_info != NULL) {
	    func_list->get_file_info(dibitmap->handler->instance, &dibitmap->handler->file_info);
	    if (info != NULL) {
		memcpy(info, &dibitmap->handler->file_info, sizeof(spDIBitmapInfo));
	    }
	}

	strict_mask = options & SP_DI_BITMAP_STRICT_MASK;
        spDebug(100, "spCreateDIBitmapFromFileWithOptions", "original strict_mask = %lx\n", strict_mask);
	
	if (func_list->get_info(dibitmap->handler->instance, &dibitmap->info, &strict_mask) == SP_TRUE) {
	    spDebug(80, "spCreateDIBitmapFromFileWithOptions", "caps = %lx, strict_mask = %lx\n", caps, strict_mask);
	    host_upside_down = dibitmap->info.upside_down;

	    if (!(options & SP_DI_BITMAP_OPTION_FORCE_NONNATIVE)) {
		if (caps & SP_DI_BITMAP_CAPS_READ_REPLACE_BITMAP) {
                    if (func_list->set_info != NULL) {
                        flag = func_list->set_info(dibitmap->handler->instance, &dibitmap->info);
                        spDebug(100, "spCreateDIBitmapFromFileWithOptions", "set_info result: %d\n", flag);
                    } else {
                        flag = SP_TRUE;
                    }
                    if (flag) {
                        if ((dibitmap->dibitmap = func_list->read(dibitmap->handler->instance, &dibitmap->info, NULL)) != NULL) {
                            flag = SP_TRUE;
                        } else {
                            flag = SP_FALSE;
                        }
                    }
		} else {
		    strict_mask2 = strict_mask;
		    if ((options & SP_DI_BITMAP_OPTION_ACCEPT_NONNATIVE)
			|| spUpdateDIBitmapInfoArch(&dibitmap->info, &strict_mask2) == SP_TRUE) {
			if (options & SP_DI_BITMAP_OPTION_ACCEPT_NONNATIVE) {
			    strict_mask2 |= SP_DI_BITMAP_STRICT_ALL | SP_DI_BITMAP_OPTION_ACCEPT_NONNATIVE;
			    spDebug(80, "spCreateDIBitmapFromFileWithOptions", "strict_mask2 updated: %lx\n", strict_mask2);
			}
			flag = func_list->get_info(dibitmap->handler->instance, &dibitmap->info, &strict_mask2);
			spDebug(80, "spCreateDIBitmapFromFileWithOptions",
				"before create bitmap: flag = %d, strict_mask2 = %lx, bit_size = %ld, pixel_stride = %ld\n",
				flag, strict_mask2, dibitmap->info.bit_size, dibitmap->info.pixel_stride);
		
			if (flag) {
			    dibitmap->dibitmap = spCreateDIBitmapArch(&dibitmap->info, strict_mask2);
                            spDebug(80, "spCreateDIBitmapFromFileWithOptions", "dibitmap->dibitmap = %ld\n",
                                    (long)dibitmap->dibitmap);
			    flag = SP_FALSE;
			}
		    } else {
			spDebug(10, "spCreateDIBitmapFromFileWithOptions", "spUpdateDIBitmapInfoArch failed\n");
		    }
		}
	    }
			
	    if ((options & SP_DI_BITMAP_OPTION_FORCE_NONNATIVE)
		|| (dibitmap->dibitmap == NULL
		    && (options & SP_DI_BITMAP_OPTION_ACCEPT_NONNATIVE))) {
                if (!(caps & (SP_DI_BITMAP_CAPS_READ_USE_BITMAP_ARGUMENT | SP_DI_BITMAP_CAPS_READ_REPLACE_BITMAP))) {
                    spDebug(80, "spCreateDIBitmapFromFileWithOptions", "create memory bitmap\n");
                    dibitmap->dibitmap = spAllocDIBitmapMemoryNonnative(&dibitmap->info);
                    use_nonnative = SP_TRUE;
                }
	    }

	    if (flag == SP_FALSE) {
		if (dibitmap->dibitmap != NULL) {
		    void *newbitmap;
                    
                    if (func_list->set_info != NULL) {
                        flag = func_list->set_info(dibitmap->handler->instance, &dibitmap->info);
                        spDebug(100, "spCreateDIBitmapFromFileWithOptions", "set_info result: %d\n", flag);
                    } else {
                        flag = SP_TRUE;
                    }
	    
                    if (flag) {
                        spLockDIBitmap(dibitmap);
                        spDebug(80, "spCreateDIBitmapFromFileWithOptions", "call read\n");
                        if ((newbitmap = func_list->read(dibitmap->handler->instance, &dibitmap->info, dibitmap->dibitmap)) != NULL) {
                            flag = SP_TRUE;
                        } else {
                            flag = SP_FALSE;
                        }
                        spUnlockDIBitmap(dibitmap);

                        if (newbitmap != NULL && newbitmap != dibitmap->dibitmap) {
                            if (use_nonnative == SP_TRUE) {
                                xfree(dibitmap->dibitmap);
                            } else {
                                spDestroyDIBitmapArch(dibitmap->dibitmap);
                            }
                            dibitmap->dibitmap = newbitmap;
                        }
                    }
		} else if (use_nonnative == SP_FALSE) {
		    if (func_list->read_rgba != NULL) {
			spDebug(80, "spCreateDIBitmapFromFileWithOptions", "create bitmap by using read_rgba\n");
			    
			if (dibitmap->info.upside_down != host_upside_down
			    || ((caps & SP_DI_BITMAP_CAPS_UPSIDE_UP) && !(caps & SP_DI_BITMAP_CAPS_UPSIDE_DOWN)
				&& dibitmap->info.upside_down != SP_FALSE)
			    || (!(caps & SP_DI_BITMAP_CAPS_UPSIDE_UP) && (caps & SP_DI_BITMAP_CAPS_UPSIDE_DOWN)
				&& dibitmap->info.upside_down == SP_FALSE)) {
			    updown_swap = SP_TRUE;
			} else {
			    updown_swap = SP_FALSE;
			}
			spDebug(80, "spCreateDIBitmapFromFileWithOptions", "updown_swap = %d\n", updown_swap);
			    
			if ((dibitmap->dibitmap = spCreateDIBitmapArch(&dibitmap->info, options)) != NULL) {
			    flag = SP_TRUE;
			    for (y = 0; y < dibitmap->info.height; y++) {
				for (x = 0; x < dibitmap->info.width; x++) {
				    r = g = b = 0; a = 0xff;
				    flag = func_list->read_rgba(dibitmap->handler->instance, &r, &g, &b, &a);
				    if (flag == SP_FALSE) {
					break;
				    }
				    if (updown_swap) {
					spSetDIBDataRGBAArch(&dibitmap->info, dibitmap->dibitmap, x, dibitmap->info.height - 1 - y, r, g, b, a);
				    } else {
					spSetDIBDataRGBAArch(&dibitmap->info, dibitmap->dibitmap, x, y, r, g, b, a);
				    }
				}
				if (flag == SP_FALSE) {
				    break;
				}
			    }
			    spDebug(80, "spCreateDIBitmapFromFileWithOptions", "flag = %d\n", flag);
			} else {
			    spDebug(10, "spCreateDIBitmapFromFileWithOptions", "spCreateDIBitmapArch failed\n");
			}
		    }
		}
	    }
	} else {
	    spDebug(10, "spCreateDIBitmapFromFileWithOptions", "get_info failed\n");
	}

	if (func_list->close != NULL) {
            spDebug(80, "spCreateDIBitmapFromFileWithOptions", "call close\n");
	    func_list->close(dibitmap->handler->instance);
	}
    } else {
	spDebug(10, "spCreateDIBitmapFromFileWithOptions", "open failed\n");
    }

    spDebug(80, "spCreateDIBitmapFromFileWithOptions", "done: flag = %d\n", flag);
    
    if (flag == SP_FALSE) {
	spDestroyDIBitmap(dibitmap);
	return NULL;
    } else {
	return dibitmap;
    }
}

spDIBitmap spCreateDIBitmapFromFile(const char *filename)
{
    return spCreateDIBitmapFromFileWithOptions(filename, NULL, 0L);
}

spDIBitmap spCreateDIBitmapFromFileNonnative(const char *filename)
{
    return spCreateDIBitmapFromFileWithOptions(filename, NULL, SP_DI_BITMAP_OPTION_ACCEPT_NONNATIVE);
}

spDIBitmapInfo *spGetDIBitmapFileInfo(spDIBitmap dibitmap)
{
    if (dibitmap == NULL || dibitmap->handler == NULL || dibitmap->handler->file_info.pixel_format == 0L) return NULL;

    return &dibitmap->handler->file_info;
}

spDIBitmap spCreateDIBitmapWithOptions(spDIBitmapInfo *info, unsigned long options)
{
    spDIBitmap dibitmap;
    spDIBitmapHandler handler;
    
    if ((handler = spGetDIBitmapSupportedHandler(0L, options, NULL, NULL, NULL)) == NULL) {
	return NULL;
    }
    
    dibitmap = spCreateDIBitmapCore(info, handler, options);

    if ((dibitmap->dibitmap = spCreateDIBitmapArch(&dibitmap->info, options)) == NULL) {
	spDestroyDIBitmap(dibitmap);
	return NULL;
    }
    
    return dibitmap;
}

spDIBitmap spCreateDIBitmap(spDIBitmapInfo *info)
{
    return spCreateDIBitmapWithOptions(info, 0L);
}

spBool spSaveDIBitmapWithHandler(spDIBitmapHandler handler, spDIBitmap dibitmap, const char *filename, spDIBitmapInfo *target_info, unsigned long options)
{
    spBool flag = SP_FALSE;
    unsigned long host_caps;
    spDIBitmapFuncList *func_list;
    spDIBitmapInfoInternal info;

    if (handler == NULL || dibitmap == NULL) return SP_FALSE;
    
    func_list = handler->func_list;
    host_caps = handler->caps;
    
    if (func_list->open(handler->instance, filename, "w") == SP_TRUE) {
	flag = SP_TRUE;

	if (target_info != NULL && func_list->set_file_info != NULL) {
	    spDebug(100, "spSaveDIBitmapWithHandler", "bit_size = %ld, pixel_format = %lx\n",
		    target_info->bit_size,  target_info->pixel_format);
	    flag = func_list->set_file_info(handler->instance, target_info, &options);
	    spDebug(100, "spSaveDIBitmapWithHandler", "set_file_info result: %d\n", flag);
	}
	
	if (flag) {
	    memcpy(&info, &dibitmap->info, sizeof(spDIBitmapInfoInternal));

	    if (func_list->set_info != NULL) {
		flag = func_list->set_info(handler->instance, &info);
		spDebug(100, "spSaveDIBitmapWithHandler", "set_info result: %d\n", flag);
	    }
	}

	if (flag) {
	    spLockDIBitmap(dibitmap);
            info.data = dibitmap->info.data;
	    flag = func_list->write(handler->instance, &info, dibitmap->dibitmap);
	    spDebug(100, "spSaveDIBitmapWithHandler", "write result: %d\n", flag);
	    spUnlockDIBitmap(dibitmap);
	} else if (func_list->write_rgba != NULL) {
	    spBool updown_swap;
	    int x, y;
	    int r, g, b, a;

	    flag = SP_TRUE;
	    if (func_list->get_info != NULL) {
		options = 0L;
		flag = func_list->get_info(handler->instance, &info, &options);
		if (flag && func_list->set_info != NULL) {
		    flag = func_list->set_info(handler->instance, &info);
		}
	    }

	    if (flag) {
		if (info.upside_down != dibitmap->info.upside_down
		    || ((host_caps & SP_DI_BITMAP_CAPS_UPSIDE_UP) && !(host_caps & SP_DI_BITMAP_CAPS_UPSIDE_DOWN)
			&& dibitmap->info.upside_down != SP_FALSE)
		    || (!(host_caps & SP_DI_BITMAP_CAPS_UPSIDE_UP) && (host_caps & SP_DI_BITMAP_CAPS_UPSIDE_DOWN)
			&& dibitmap->info.upside_down == SP_FALSE)) {
		    updown_swap = SP_TRUE;
		} else {
		    updown_swap = SP_FALSE;
		}
		spDebug(80, "spSaveDIBitmapWithHandler", "updown_swap = %d\n", updown_swap);
	    
		spLockDIBitmap(dibitmap);
		for (y = 0; y < dibitmap->info.height; y++) {
		    for (x = 0; x < dibitmap->info.width; x++) {
			r = g = b = 0; a = 0xff;
			if (updown_swap) {
			    spGetDIBitmapRGBA(dibitmap, x, dibitmap->info.height - 1 - y, &r, &g, &b, &a);
			} else {
			    spGetDIBitmapRGBA(dibitmap, x, y, &r, &g, &b, &a);
			}
			
			if (!(info.pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)) {
			    a = 0xff;
			}
			spDebug(100, "spSaveDIBitmapWithHandler", "x = %d, y = %d, a = %x\n", x, y, a);
		    
			flag = func_list->write_rgba(handler->instance, r, g, b, a);
			if (flag == SP_FALSE) {
			    break;
			}
		    }
		    if (flag == SP_FALSE) {
			break;
		    }
		}
		spUnlockDIBitmap(dibitmap);
	    } else {
		spDebug(10, "spSaveDIBitmapWithHandler", "get_info failed\n");
	    }
	}
	
	if (func_list->close != NULL) {
	    func_list->close(handler->instance);
	}
    } else {
	spDebug(10, "spSaveDIBitmapWithHandler", "open failed\n");
    }

    spDebug(100, "spSaveDIBitmapWithHandler", "done: %d\n", flag);

    return flag;

}

spBool spSaveDIBitmapWithInfo(spDIBitmap dibitmap, const char *filename, spDIBitmapInfo *target_info, unsigned long options)
{
    spBool flag = SP_FALSE;
    unsigned long caps;
    spDIBitmapHandler handler;

    if (dibitmap == NULL) return SP_FALSE;

    caps = spCalcDIBitmapRequiredCaps(&dibitmap->info, options);

    if (spIsDIBitmapSupportedByHandler(dibitmap->handler, caps, options, filename, "w") == SP_TRUE) {
	spDebug(80, "spSaveDIBitmapWithInfo", "use current handler\n");
	handler = dibitmap->handler;
    } else {
	if (dibitmap->info.pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK) {
	    options |= SP_DI_BITMAP_OPTION_AVOID_NATIVE_BASED_IO;
	}
	if ((handler = spGetDIBitmapSupportedHandler(caps, options, filename, "w", NULL)) == NULL) {
	    spDebug(10, "spSaveDIBitmapWithInfo", "spGetDIBitmapSupportedHandler failed\n");
	    return SP_FALSE;
	}
	spDebug(80, "spSaveDIBitmapWithInfo", "use new handler\n");
    }

    flag = spSaveDIBitmapWithHandler(handler, dibitmap, filename, target_info, options);

    if (handler != dibitmap->handler) {
	freeDIBitmapHandler(handler);
    }
    
    spDebug(100, "spSaveDIBitmapWithInfo", "done: %d\n", flag);
    
    return flag;
}

spBool spSaveDIBitmap(spDIBitmap dibitmap, const char *filename)
{
    return spSaveDIBitmapWithInfo(dibitmap, filename, NULL, 0L);
}

spBool spDestroyDIBitmap(spDIBitmap dibitmap)
{
    if (dibitmap == NULL) return SP_FALSE;

    if (dibitmap->dibitmap != NULL) {
	if (dibitmap->info.pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK) {
            xfree(dibitmap->dibitmap);
        } else {
            spDestroyDIBitmapArch(dibitmap->dibitmap);
        }
    }

    if (dibitmap->handler != NULL) {
	freeDIBitmapHandler(dibitmap->handler);
    }

    xfree(dibitmap);
    
    return SP_TRUE;
}

unsigned long spGetDIBitmapCaps(spDIBitmap dibitmap)
{
    if (dibitmap == NULL) return 0L;

    return dibitmap->handler->func_list->get_caps(dibitmap->handler->instance);
}

spDIPixelFormat spGetDIBitmapPixelFormat(spDIBitmap dibitmap)
{
    if (dibitmap == NULL) return 0L;

    return SP_DI_PIXEL_FORMAT_MASK & dibitmap->info.pixel_format;
}

spBool spIsDIBitmapTypePlugin(spDIBitmap dibitmap)
{
    if (dibitmap == NULL) return SP_FALSE;

    if (SP_DI_BITMAP_TYPE_PLUGIN_MASK & dibitmap->info.pixel_format) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spLockDIBitmap(spDIBitmap dibitmap)
{
    if (dibitmap->info.locked == SP_TRUE) {
	return SP_FALSE;
    }
    
    dibitmap->info.locked = SP_TRUE;

    if (dibitmap->info.pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK) {
	dibitmap->info.data = dibitmap->dibitmap;
	return SP_TRUE;
    } else {
	return spLockDIBitmapArch(&dibitmap->info, dibitmap->dibitmap);
    }
}

spBool spUnlockDIBitmap(spDIBitmap dibitmap)
{
    if (dibitmap->info.locked == SP_FALSE) {
	return SP_FALSE;
    }
    
    dibitmap->info.locked = SP_FALSE;
    
    if (dibitmap->info.pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK) {
	dibitmap->info.data = NULL;
	return SP_TRUE;
    } else {
	return spUnlockDIBitmapArch(&dibitmap->info, dibitmap->dibitmap);
    }
}

static unsigned char *getDIBitmapDataPtr(spDIBitmapInfoInternal *info, unsigned char *data, int x, int y)
{
    int pos;

    if (info->upside_down == SP_TRUE) {
	pos = (info->height - y - 1) * info->line_stride + x * info->pixel_stride;
    } else {
	pos = y * info->line_stride + x * info->pixel_stride;
    }
    if (info->pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK) {
	pos += (info->bit_size / 8);
    }
    
    return (unsigned char *)(data + pos);
}

spBool spSetDIBDataRGBADefault(spDIBitmapInfoInternal *info, void *dibdata, int x, int y, int r, int g, int b, int a)
{
    unsigned char *data;

    data = getDIBitmapDataPtr(info, info->data, x, y);
    
    if (info->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK) {
	data[2] = r; data[1] = g; data[0] = b;
    } else {
	data[0] = r; data[1] = g; data[2] = b;
    }

    return SP_TRUE;
}

spBool spAddDIBDataRGBADefault(spDIBitmapInfoInternal *info, void *dibdata, int x, int y, int r, int g, int b, int a)
{
    unsigned char *data;
    int r2, g2, b2;

    data = getDIBitmapDataPtr(info, info->data, x, y);
    spDebug(100, "spAddDIBDataRGBADefault", "input (%d, %d): %x %x %x\n", x, y, data[0], data[1], data[2]);
    
    if (info->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK) {
	r2 = MIN(MAX(data[2] + r, 0), 255);
	data[2] = r2;
	g2 = MIN(MAX(data[1] + g, 0), 255);
	data[1] = g2;
	b2 = MIN(MAX(data[0] + b, 0), 255);
	data[0] = b2;
    } else {
	r2 = MIN(MAX(data[0] + r, 0), 255);
	data[0] = r2;
	g2 = MIN(MAX(data[1] + g, 0), 255);
	data[1] = g2;
	b2 = MIN(MAX(data[2] + b, 0), 255);
	data[2] = b2;
    }

    spDebug(100, "spAddDIBDataRGBADefault", "output: %x %x %x\n", data[0], data[1], data[2]);
    
    return SP_TRUE;
}

spBool spGetDIBDataRGBADefault(spDIBitmapInfoInternal *info, void *dibdata, int x, int y, int *r, int *g, int *b, int *a)
{
    unsigned char *data;

    data = getDIBitmapDataPtr(info, info->data, x, y);
    spDebug(100, "spGetDIBDataRGBADefault", "%x %x %x\n", data[0], data[1], data[2]);
    
    if (info->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK) {
	*r = data[2]; *g = data[1]; *b = data[0];
    } else {
	*r = data[0]; *g = data[1]; *b = data[2];
    }

    if (a != NULL) {
	if (!(info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)) {
	    *a = 255;
	} else {
	    if (info->pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK) {
		*a = *(data - 1);
	    } else {
		*a = data[3];
	    }
	}
    }

    return SP_TRUE;
}

#if !defined(USE_MOTIF)
spBool spSetDIBDataRGBAArch(spDIBitmapInfoInternal *info, void *dibdata, int x, int y, int r, int g, int b, int a)
{
    return spSetDIBDataRGBADefault(info, dibdata, x, y, r, g, b, a);
}
spBool spAddDIBDataRGBAArch(spDIBitmapInfoInternal *info, void *dibdata, int x, int y, int r, int g, int b, int a)
{
    return spAddDIBDataRGBADefault(info, dibdata, x, y, r, g, b, a);
}
spBool spGetDIBDataRGBAArch(spDIBitmapInfoInternal *info, void *dibdata, int x, int y, int *r, int *g, int *b, int *a)
{
    return spGetDIBDataRGBADefault(info, dibdata, x, y, r, g, b, a);
}
#endif

spBool spSetDIBitmapRGBA(spDIBitmap dibitmap, int x, int y, int r, int g, int b, int a)
{
    if (dibitmap->info.locked == SP_FALSE || dibitmap->info.data == NULL) {
	return SP_FALSE;
    }
    if (!(dibitmap->info.pixel_format & (SP_DI_PIXEL_FORMAT_RGB_MASK))) {
	return SP_FALSE;
    }

    if (dibitmap->info.pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK) {
	return spSetDIBDataRGBADefault(&dibitmap->info, dibitmap->dibitmap, x, y, r, g, b, a);
    } else {
	return spSetDIBDataRGBAArch(&dibitmap->info, dibitmap->dibitmap, x, y, r, g, b, a);
    }
}

spBool spSetDIBitmapRGB(spDIBitmap dibitmap, int x, int y, int r, int g, int b)
{
    if (dibitmap->info.locked == SP_FALSE || dibitmap->info.data == NULL) {
	return SP_FALSE;
    }
    if (!(dibitmap->info.pixel_format & SP_DI_PIXEL_FORMAT_RGB_MASK)) {
	return SP_FALSE;
    }
    
    if (dibitmap->info.pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK) {
	return spSetDIBDataRGBADefault(&dibitmap->info, dibitmap->dibitmap, x, y, r, g, b, 255);
    } else {
	return spSetDIBDataRGBAArch(&dibitmap->info, dibitmap->dibitmap, x, y, r, g, b, 255);
    }
}

/* This function includes clipping procedure (limiting to 0-255) while spSetDIBitmapRGB doesn't. */
spBool spAddDIBitmapRGBA(spDIBitmap dibitmap, int x, int y, int r, int g, int b, int a)
{
    if (dibitmap->info.locked == SP_FALSE || dibitmap->info.data == NULL) {
	return SP_FALSE;
    }
    if (!(dibitmap->info.pixel_format & (SP_DI_PIXEL_FORMAT_RGB_MASK))) {
	return SP_FALSE;
    }

    if (dibitmap->info.pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK) {
	return spAddDIBDataRGBADefault(&dibitmap->info, dibitmap->dibitmap, x, y, r, g, b, a);
    } else {
	return spAddDIBDataRGBAArch(&dibitmap->info, dibitmap->dibitmap, x, y, r, g, b, a);
    }
}

spBool spAddDIBitmapRGB(spDIBitmap dibitmap, int x, int y, int r, int g, int b)
{
    if (dibitmap->info.locked == SP_FALSE || dibitmap->info.data == NULL) {
	return SP_FALSE;
    }
    if (!(dibitmap->info.pixel_format & SP_DI_PIXEL_FORMAT_RGB_MASK)) {
	return SP_FALSE;
    }

    if (dibitmap->info.pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK) {
	return spAddDIBDataRGBADefault(&dibitmap->info, dibitmap->dibitmap, x, y, r, g, b, 255);
    } else {
	return spAddDIBDataRGBAArch(&dibitmap->info, dibitmap->dibitmap, x, y, r, g, b, 255);
    }
}

spBool spGetDIBitmapRGBA(spDIBitmap dibitmap, int x, int y, int *r, int *g, int *b, int *a)
{
    if (dibitmap->info.locked == SP_FALSE || dibitmap->info.data == NULL) {
	spDebug(100, "spGetDIBDataRGBA", "not locked or no data\n");
	return SP_FALSE;
    }
    if (!(dibitmap->info.pixel_format & (SP_DI_PIXEL_FORMAT_RGB_MASK))) {
	spDebug(100, "spGetDIBDataRGBA", "pixel format is not rgba\n");
	return SP_FALSE;
    }

    if (dibitmap->info.pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK) {
	return spGetDIBDataRGBADefault(&dibitmap->info, dibitmap->dibitmap, x, y, r, g, b, a);
    } else {
	return spGetDIBDataRGBAArch(&dibitmap->info, dibitmap->dibitmap, x, y, r, g, b, a);
    }
}

spBool spGetDIBitmapRGB(spDIBitmap dibitmap, int x, int y, int *r, int *g, int *b)
{
    if (dibitmap->info.locked == SP_FALSE || dibitmap->info.data == NULL) {
	spDebug(100, "spGetDIBDataRGB", "not locked or no data\n");
	return SP_FALSE;
    }
    if (!(dibitmap->info.pixel_format & SP_DI_PIXEL_FORMAT_RGB_MASK)) {
	spDebug(100, "spGetDIBDataRGB", "pixel format is not rgb\n");
	return SP_FALSE;
    }

    if (dibitmap->info.pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK) {
	return spGetDIBDataRGBADefault(&dibitmap->info, dibitmap->dibitmap, x, y, r, g, b, NULL);
    } else {
	return spGetDIBDataRGBAArch(&dibitmap->info, dibitmap->dibitmap, x, y, r, g, b, NULL);
    }
}

#if defined(GTK) || defined(USE_MOTIF)
spBool spStretchDIBitmapToImageArch(spDIBitmapInfoInternal *info, void *bitmap, spComponent dest,
				    int src_x, int src_y, int src_width, int src_height,
				    int dest_x, int dest_y, int dest_width, int dest_height)
{
    double xzoom, yzoom;
    double dest_xd, dest_yd;
    int dest_px, dest_py;
    int point_width, point_height;
    int x, y;
    int src_left, src_bottom;
    int dest_x2, dest_y2;
    int r, g, b;
    spGraphics graphics;
    spPixel pixel;

    if ((graphics = spCreateGraphics("_spDIBitmapStretchImage", NULL)) == NULL) {
	return SP_FALSE;
    }

    src_width = MIN(info->width - src_x, src_width);
    src_height = MIN(info->height - src_y, src_height);
	
    xzoom = (double)dest_width / (double)src_width;
    yzoom = (double)dest_height / (double)src_height;
    spDebug(100, "spStretchDIBitmapToImageArch", "xzoom = %f, yzoom = %f\n", xzoom, yzoom);

    src_left = src_x + src_width;
    src_bottom = src_y + src_height;

    dest_py = dest_y;
    dest_yd = (double)dest_y + yzoom;

    for (y = src_y + 1; y <= src_bottom; y++) {
	spDebug(100, "spStretchDIBitmapToImageArch", "y = %d / %d, dest_yd = %f\n",
		y, src_bottom, dest_yd);
	
	dest_px = dest_x;
	dest_xd = (double)dest_x + xzoom;

	if (y == src_bottom) {
	    dest_y2 = dest_height;
	} else {
	    dest_y2 = spRound(dest_yd);
	}
	point_height = dest_y2 - dest_py;
	spDebug(100, "spStretchDIBitmapToImageArch", "point_height = %d, dest_y2 = %d, dest_py = %d\n",
		point_height, dest_y2, dest_py);
	
	if (point_height >= 1) {
	    for (x = src_x + 1; x <= src_left; x++) {
		spDebug(100, "spStretchDIBitmapToImageArch", "x = %d / %d, dest_xd = %f\n",
			x, src_left, dest_xd);

		if (x == src_left) {
		    dest_x2 = dest_width;
		} else {
		    dest_x2 = spRound(dest_xd);
		}
		point_width = dest_x2 - dest_px;
		spDebug(100, "spStretchDIBitmapToImageArch", "point_width = %d, dest_x2 = %d, dest_px = %d\n",
			point_width, dest_x2, dest_px);
		
		if (point_width >= 1) {
		    spGetDIBDataRGBAArch(info, bitmap, x - 1, y - 1, &r, &g, &b, NULL);
		    spDebug(100, "spStretchDIBitmapToImageArch", "rgb = %d %d %d\n", r, g, b);
		    
		    pixel = spRGB(r, g, b);
#if 0
		    spSetGraphicsParams(graphics,
					SppForegroundPixel, pixel,
					NULL);
#else
		    spSetForegroundPixel(graphics, pixel);
#endif
		    spDebug(100, "spStretchDIBitmapToImageArch", "before draw, point_width = %d, point_height = %d\n",
			    point_width, point_height);
		    
		    if (point_width == 1 && point_height == 1) {
			spDrawPoint(dest, graphics, dest_px, dest_py);
		    } else {
			spFillRectangle(dest, graphics, dest_px, dest_py, point_width, point_height);
		    }
	    
		    spDebug(100, "spStretchDIBitmapToImageArch", "after draw, dest_px = %d, dest_x2 = %d\n", dest_px, dest_x2);
		    dest_px = dest_x2;
		}
	    
		dest_xd += xzoom;
	    }
	    
	    dest_py = dest_y2;
	}
	
	dest_yd += yzoom;
    }
	
    spFreeGraphics(graphics);

    return SP_TRUE;
}
#endif

spBool spCopyDIBitmapToImage(spDIBitmap dibitmap, spComponent dest, int src_x, int src_y,
			     int width, int height, int dest_x, int dest_y)
{
    if (spIsDrawable(dest) == SP_FALSE || dibitmap->info.locked == SP_FALSE
	|| (dibitmap->info.pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK)) return SP_FALSE;

    return spCopyDIBitmapToImageArch(&dibitmap->info, dibitmap->dibitmap, dest, src_x, src_y,
				     width, height, dest_x, dest_y);
}

spBool spStretchDIBitmapToImage(spDIBitmap dibitmap, spComponent dest,
				int src_x, int src_y, int src_width, int src_height,
				int dest_x, int dest_y, int dest_width, int dest_height)
{
    if (src_width == dest_width && src_height == dest_height) {
	spDebug(100, "spStretchDIBitmapToImage", "call spCopyDIBitmapToImage\n");
	return spCopyDIBitmapToImage(dibitmap, dest, src_x, src_y,
				     src_width, src_height,  dest_x, dest_y);
    }
    
    if (spIsDrawable(dest) == SP_FALSE || dibitmap->info.locked == SP_FALSE
	|| (dibitmap->info.pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK)) {
        spDebug(100, "spStretchDIBitmapToImage", "not drawable\n");
        return SP_FALSE;
    }

    return spStretchDIBitmapToImageArch(&dibitmap->info, dibitmap->dibitmap, dest,
					src_x, src_y, src_width, src_height,
					dest_x, dest_y, dest_width, dest_height);
}

spBool spCopyImageToDIBitmap(spComponent src, spDIBitmap dibitmap, int src_x, int src_y,
			     int width, int height, int dest_x, int dest_y)
{
    if (spIsDrawable(src) == SP_FALSE || dibitmap->info.locked == SP_FALSE
	|| (dibitmap->info.pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK)) return SP_FALSE;

    dibitmap->dibitmap = spCopyImageToDIBitmapArch(src, &dibitmap->info, dibitmap->dibitmap, src_x, src_y,
						   width, height, dest_x, dest_y);
    if (dibitmap->dibitmap != NULL) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spIsDIBitmapPlugin(spPlugin *plugin)
{
    if (plugin == NULL) return SP_FALSE;

    if (spEqPluginType(plugin, SP_PLUGIN_DIBITMAP) == SP_TRUE) {
	spDebug(100, "spIsDIBitmapPlugin", "SP_TRUE\n");
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

long spReadBmpFileHeader(DWORD *pbfSize, DWORD *pbfOffBits, FILE *fp)
{
    char type[2];
    DWORD bfSize, bfOffBits;
    WORD bfReserved;
    
    spDebug(50, "spReadBmpFileHeader", "in\n");
    
    spSeekFile(fp, 0L, SEEK_SET);

    if (fread(type, 1, 2, fp) != 2 || !strneq(type, "BM", 2)) {
        return 0;
    }
    if (spFReadDWORD(&bfSize, 1, fp) != 1) {
        return 0;
    }
    if (pbfSize != NULL) *pbfSize = bfSize;
    
    if (spFReadWORD(&bfReserved, 1, fp) != 1) {
        return 0;
    }
    if (spFReadWORD(&bfReserved, 1, fp) != 1) {
        return 0;
    }
    if (spFReadDWORD(&bfOffBits, 1, fp) != 1) {
        return 0;
    }
    if (pbfSize != NULL) *pbfOffBits = bfOffBits;

    spDebug(50, "spReadBmpFileHeader", "bfSize = %ld, bfOffBits = %ld\n", bfSize, bfOffBits);
    
    return 14;
}

/* reference: http://www.ruche-home.net/program/bmp/bitfield */
spBool spCheckBmpBitFieldMasks(DWORD *masks)
{
    int i;
    spLong32 n, mask;
    spLong32 allMask = 0;
 
    for (i = 0; i < 3; ++i)
    {
        mask = (spLong32)masks[i];
        n = mask + (mask & -mask);
        if ((n & (n - 1)) != 0) {
            /* bits are not sequencial */
            return SP_FALSE;
        }
 
        if ((allMask & mask) != 0) {
            /* bits overlap */
            return SP_FALSE;
        }
        allMask |= mask;
    }
 
    return SP_TRUE;
}

long spReadBmpBitFields(DWORD *masks, spBool *valid_flag, FILE *fp)
{
    int i;
    long read_size;

    for (i = 0, read_size = 0; i < 3; i++) {
        if (spFReadDWORD(&masks[i], 1, fp) <= 0) {
            if (valid_flag != NULL) *valid_flag = SP_FALSE;
            return read_size;
        }
        read_size += 4;
    }

    if (valid_flag != NULL) {
        *valid_flag = spCheckBmpBitFieldMasks(masks);
    }

    return read_size;
}

long spReadBmpV5Header(BITMAPV5HEADER *header, DWORD bfOffBits, FILE *fp)
{
    long read_size;
    long remain_size;

    spDebug(50, "spReadBmpV5Header", "in: sizeof(BITMAPV5HEADER) = %d, sizeof(long) = %d, sizeof(header->bV5Size) = %d\n",
            sizeof(BITMAPV5HEADER), sizeof(long), sizeof(header->bV5Size));
    
    memset(header, 0, sizeof(BITMAPV5HEADER));

    if (spFReadDWORD(&header->bV5Size, 1, fp) <= 0) {
        spDebug(10, "spReadBmpV5Header", "spFReadDWORD failed\n");
	return 0;
    }
    read_size = 4;

    spDebug(50, "spReadBmpV5Header", "header->bV5Size = %ld\n", header->bV5Size);
    
    if (header->bV5Size <= 12) {
        spWORD value;
        if (spFReadWORD(&value, 1, fp) <= 0) {
            return 0;
        }
        header->bV5Width = value;
        read_size += 2;

        if (spFReadWORD(&value, 1, fp) <= 0) {
            return 0;
        }
        header->bV5Height = value;
        read_size += 2;
    } else {
        if (spFReadLONG(&header->bV5Width, 1, fp) <= 0) {
            return 0;
        }
        read_size += 4;

        if (spFReadLONG(&header->bV5Height, 1, fp) <= 0) {
            return 0;
        }
        read_size += 4;
    }

    if (spFReadWORD(&header->bV5Planes, 1, fp) <= 0) {
	return 0;
    }
    read_size += 2;

    if (spFReadWORD(&header->bV5BitCount, 1, fp) <= 0) {
	return 0;
    }
    read_size += 2; /* 12 or 16 bytes */

    spDebug(50, "spReadBmpV5Header",
	    "bV5Size = %ld, bV5Width = %ld, bV5Height = %ld, bV5Planes = %d, bV5BitCount = %d\n",
	    header->bV5Size, header->bV5Width, header->bV5Height, header->bV5Planes, header->bV5BitCount);
    
    if ((long)header->bV5Size > read_size) {
        if (spFReadDWORD(&header->bV5Compression, 1, fp) <= 0) {
            return 0;
        }
        read_size += 4;

        if (spFReadDWORD(&header->bV5SizeImage, 1, fp) <= 0) {
            return 0;
        }
        read_size += 4;

        if (spFReadLONG(&header->bV5XPelsPerMeter, 1, fp) <= 0) {
            return 0;
        }
        read_size += 4;

        if (spFReadLONG(&header->bV5YPelsPerMeter, 1, fp) <= 0) {
            return 0;
        }
        read_size += 4;

        if (spFReadDWORD(&header->bV5ClrUsed, 1, fp) <= 0) {
            return 0;
        }
        read_size += 4;

        if (spFReadDWORD(&header->bV5ClrImportant, 1, fp) <= 0) {
            return 0;
        }
        read_size += 4; /* 40 bytes */

        spDebug(50, "spReadBmpV5Header",
                "bV5Compression = %ld, bV5SizeImage = %ld, bV5XPelsPerMeter = %ld, bV5YPelsPerMeter = %ld, bV5ClrUsed = %ld, bV5ClrImportant = %ld\n",
                header->bV5Compression, header->bV5SizeImage, header->bV5XPelsPerMeter, header->bV5YPelsPerMeter,
                header->bV5ClrUsed, header->bV5ClrImportant);
    
        if ((long)header->bV5Size > read_size) {
            if (spFReadDWORD(&header->bV5RedMask, 1, fp) <= 0) {
                return 0;
            }
            read_size += 4;

            if (spFReadDWORD(&header->bV5GreenMask, 1, fp) <= 0) {
                return 0;
            }
            read_size += 4;

            if (spFReadDWORD(&header->bV5BlueMask, 1, fp) <= 0) {
                return 0;
            }
            read_size += 4; /* 52 bytes */

            if ((long)header->bV5Size > read_size) {
                if (spFReadDWORD(&header->bV5AlphaMask, 1, fp) <= 0) {
                    return 0;
                }
                read_size += 4; /* 56 bytes */

                spDebug(50, "spReadBmpV5Header",
                        "bV5RedMask = %lx, bV5GreenMask = %lx, bV5BlueMask = %lx, bV5AlphaMask = %lx\n",
                        header->bV5RedMask, header->bV5GreenMask, header->bV5BlueMask, header->bV5AlphaMask);
    
                if (spFReadDWORD(&header->bV5CSType, 1, fp) <= 0) {
                    return 0;
                }
                read_size += 4; /* 60 bytes */

                if ((long)header->bV5Size > read_size) {
                    if (spFReadLONG(&header->bV5Endpoints.ciexyzRed.ciexyzX, 1, fp) <= 0) {
                        return 0;
                    }
                    read_size += 4;

                    if (spFReadLONG(&header->bV5Endpoints.ciexyzRed.ciexyzY, 1, fp) <= 0) {
                        return 0;
                    }
                    read_size += 4;

                    if (spFReadLONG(&header->bV5Endpoints.ciexyzRed.ciexyzZ, 1, fp) <= 0) {
                        return 0;
                    }
                    read_size += 4;

                    if (spFReadLONG(&header->bV5Endpoints.ciexyzGreen.ciexyzX, 1, fp) <= 0) {
                        return 0;
                    }
                    read_size += 4;

                    if (spFReadLONG(&header->bV5Endpoints.ciexyzGreen.ciexyzY, 1, fp) <= 0) {
                        return 0;
                    }
                    read_size += 4;

                    if (spFReadLONG(&header->bV5Endpoints.ciexyzGreen.ciexyzZ, 1, fp) <= 0) {
                        return 0;
                    }
                    read_size += 4;

                    if (spFReadLONG(&header->bV5Endpoints.ciexyzBlue.ciexyzX, 1, fp) <= 0) {
                        return 0;
                    }
                    read_size += 4;

                    if (spFReadLONG(&header->bV5Endpoints.ciexyzBlue.ciexyzY, 1, fp) <= 0) {
                        return 0;
                    }
                    read_size += 4;

                    if (spFReadLONG(&header->bV5Endpoints.ciexyzBlue.ciexyzZ, 1, fp) <= 0) {
                        return 0;
                    }
                    read_size += 4; /* 96 bytes */

                    if ((long)header->bV5Size > read_size) {
                        spDebug(50, "spReadBmpV5Header",
                                "bV5CSType = %lx, header->bV5Endpoints.ciexyzRed = (%ld, %ld, %ld), header->bV5Endpoints.ciexyzGreen = (%ld, %ld, %ld), header->bV5Endpoints.ciexyzBlue = (%ld, %ld, %ld)\n",
                                header->bV5CSType,
                                header->bV5Endpoints.ciexyzRed.ciexyzX, header->bV5Endpoints.ciexyzRed.ciexyzY, header->bV5Endpoints.ciexyzRed.ciexyzZ,
                                header->bV5Endpoints.ciexyzGreen.ciexyzX, header->bV5Endpoints.ciexyzGreen.ciexyzY, header->bV5Endpoints.ciexyzGreen.ciexyzZ,
                                header->bV5Endpoints.ciexyzBlue.ciexyzX, header->bV5Endpoints.ciexyzBlue.ciexyzY, header->bV5Endpoints.ciexyzBlue.ciexyzZ);

                        if (spFReadDWORD(&header->bV5GammaRed, 1, fp) <= 0) {
                            return 0;
                        }
                        read_size += 4;

                        if (spFReadDWORD(&header->bV5GammaGreen, 1, fp) <= 0) {
                            return 0;
                        }
                        read_size += 4;

                        if (spFReadDWORD(&header->bV5GammaBlue, 1, fp) <= 0) {
                            return 0;
                        }
                        read_size += 4; /* 108 bytes */

                        spDebug(50, "spReadBmpV5Header", "bV5Gamma = (%ld, %ld, %ld)\n",
                                header->bV5GammaRed, header->bV5GammaGreen, header->bV5GammaBlue);
        
                        if ((long)header->bV5Size > read_size) {
                            if (spFReadDWORD(&header->bV5Intent, 1, fp) <= 0) {
                                return 0;
                            }
                            read_size += 4; /* 112 bytes */

                            if ((long)header->bV5Size > read_size) {
                                if (spFReadDWORD(&header->bV5ProfileData, 1, fp) <= 0) {
                                    return 0;
                                }
                                read_size += 4;

                                if (spFReadDWORD(&header->bV5ProfileSize, 1, fp) <= 0) {
                                    return 0;
                                }
                                read_size += 4; /* 120 bytes */

                                if ((long)header->bV5Size > read_size) {
                                    if (spFReadDWORD(&header->bV5Reserved, 1, fp) <= 0) {
                                        return 0;
                                    }
                                    read_size += 4; /* 124 bytes */
                                }
                            }
                        
                            spDebug(50, "spReadBmpV5Header", "bV5Intent = %ld, bV5ProfileData = %ld, bV5ProfileSize = %ld, bV5Reserved = %ld\n",
                                    header->bV5Intent, header->bV5ProfileData, header->bV5ProfileSize, header->bV5Reserved);
                        }
                    }
                }
            }
        }
    }

    if (bfOffBits > 0) {
        remain_size = (long)bfOffBits - (14 + read_size);
        spDebug(50, "spReadBmpV5Header",
                "remain_size = %ld, header->bV5BitCount = %d, header->bV5Compression = %ld\n",
                remain_size, header->bV5BitCount, header->bV5Compression);

        if (remain_size >= 12
            && (header->bV5BitCount == 16 || header->bV5BitCount == 32)
            && (header->bV5Compression == 0 || header->bV5Compression == 3)
            && header->bV5RedMask == 0 && header->bV5GreenMask == 0 && header->bV5BlueMask == 0) {
            long nread;
            spBool valid_flag;
            DWORD masks[3];
            nread = spReadBmpBitFields(masks, &valid_flag, fp);
            if (nread > 0 && valid_flag) {
                header->bV5RedMask = masks[0];
                header->bV5GreenMask = masks[1];
                header->bV5BlueMask = masks[2];
            }
            spDebug(50, "spReadBmpV5Header",
                    "mask after header: bV5RedMask = %lx, bV5GreenMask = %lx, bV5BlueMask = %lx, bV5AlphaMask = %lx\n",
                    header->bV5RedMask, header->bV5GreenMask, header->bV5BlueMask, header->bV5AlphaMask);
            read_size += nread;
            remain_size -= nread;
        }

        if (remain_size > 0) {
            if (fseek(fp, remain_size, SEEK_CUR) == 0) {
                read_size += remain_size;
            }
        }
    }

    if (header->bV5BitCount == 16 && header->bV5RedMask == 0 && header->bV5GreenMask == 0
        && header->bV5BlueMask == 0) {
        header->bV5RedMask = 0x00007c00;
        header->bV5GreenMask = 0x000003e0;
        header->bV5BlueMask = 0x0000001f;
    }
	    
    spDebug(50, "spReadBmpV5Header", "read_size = %ld\n", read_size);
    
    return read_size;
}

DWORD spCalcFileSizeFromBmpV5Header(BITMAPV5HEADER *header, DWORD *pbfOffBits)
{
    DWORD bfSize, bfOffBits;

    bfOffBits = 14 + header->bV5Size;

    if ((header->bV5BitCount == 16 || header->bV5BitCount == 32)
        && header->bV5Compression == 3) {
        bfOffBits += 12;
    }

    bfSize = bfOffBits + header->bV5SizeImage;

    if (pbfOffBits != NULL) *pbfOffBits = bfOffBits;

    return bfSize;
}

long spWriteBmpFileHeader(DWORD bfSize, DWORD bfOffBits, FILE *fp)
{
    char type[2];
    WORD bfReserved;
    
    spSeekFile(fp, 0L, SEEK_SET);

    type[0] = 'B'; type[1] = 'M';
    if (fwrite(type, 1, 2, fp) != 2) {
        return 0;
    }
    if (spFWriteDWORD(&bfSize, 1, fp) != 1) {
        return 0;
    }

    bfReserved = 0;
    if (spFWriteWORD(&bfReserved, 1, fp) != 1) {
        return 0;
    }
    bfReserved = 0;
    if (spFWriteWORD(&bfReserved, 1, fp) != 1) {
        return 0;
    }
    if (spFWriteDWORD(&bfOffBits, 1, fp) != 1) {
        return 0;
    }

    return 14;
}

long spWriteBmpV5Header(BITMAPV5HEADER *header, DWORD bfOffBits, FILE *fp)
{
    unsigned char c;
    long remain_size;
    long write_size;

    if (spFWriteDWORD(&header->bV5Size, 1, fp) <= 0) {
	return 0;
    }
    write_size = 4;

    if (header->bV5Size <= 12) {
        spWORD value;
        value = (short)header->bV5Width;
        if (spFWriteWORD(&value, 1, fp) <= 0) {
            return 0;
        }
        write_size += 2;

        value = (short)header->bV5Height;
        if (spFWriteWORD(&value, 1, fp) <= 0) {
            return 0;
        }
        write_size += 2;
    } else {
        if (spFWriteLONG(&header->bV5Width, 1, fp) <= 0) {
            return 0;
        }
        write_size += 4;

        if (spFWriteLONG(&header->bV5Height, 1, fp) <= 0) {
            return 0;
        }
        write_size += 4;
    }

    if (spFWriteWORD(&header->bV5Planes, 1, fp) <= 0) {
	return 0;
    }
    write_size += 2;

    if (spFWriteWORD(&header->bV5BitCount, 1, fp) <= 0) {
	return 0;
    }
    write_size += 2; /* 12 or 16 bytes */

    if ((long)header->bV5Size > write_size) {
        if (spFWriteDWORD(&header->bV5Compression, 1, fp) <= 0) {
            return 0;
        }
        write_size += 4;

        if (spFWriteDWORD(&header->bV5SizeImage, 1, fp) <= 0) {
            return 0;
        }
        write_size += 4;

        if (spFWriteLONG(&header->bV5XPelsPerMeter, 1, fp) <= 0) {
            return 0;
        }
        write_size += 4;

        if (spFWriteLONG(&header->bV5YPelsPerMeter, 1, fp) <= 0) {
            return 0;
        }
        write_size += 4;

        if (spFWriteDWORD(&header->bV5ClrUsed, 1, fp) <= 0) {
            return 0;
        }
        write_size += 4;

        if (spFWriteDWORD(&header->bV5ClrImportant, 1, fp) <= 0) {
            return 0;
        }
        write_size += 4; /* 40 bytes */

        if ((long)header->bV5Size > write_size) {
            if (spFWriteDWORD(&header->bV5RedMask, 1, fp) <= 0) {
                return 0;
            }
            write_size += 4;

            if (spFWriteDWORD(&header->bV5GreenMask, 1, fp) <= 0) {
                return 0;
            }
            write_size += 4;

            if (spFWriteDWORD(&header->bV5BlueMask, 1, fp) <= 0) {
                return 0;
            }
            write_size += 4; /* 52 bytes */

            if ((long)header->bV5Size > write_size) {
                if (spFWriteDWORD(&header->bV5AlphaMask, 1, fp) <= 0) {
                    return 0;
                }
                write_size += 4; /* 56 bytes */

                if (spFWriteDWORD(&header->bV5CSType, 1, fp) <= 0) {
                    return 0;
                }
                write_size += 4; /* 60 bytes */

                if ((long)header->bV5Size > write_size) {
                    if (spFWriteLONG(&header->bV5Endpoints.ciexyzRed.ciexyzX, 1, fp) <= 0) {
                        return 0;
                    }
                    write_size += 4;

                    if (spFWriteLONG(&header->bV5Endpoints.ciexyzRed.ciexyzY, 1, fp) <= 0) {
                        return 0;
                    }
                    write_size += 4;

                    if (spFWriteLONG(&header->bV5Endpoints.ciexyzRed.ciexyzZ, 1, fp) <= 0) {
                        return 0;
                    }
                    write_size += 4;

                    if (spFWriteLONG(&header->bV5Endpoints.ciexyzGreen.ciexyzX, 1, fp) <= 0) {
                        return 0;
                    }
                    write_size += 4;

                    if (spFWriteLONG(&header->bV5Endpoints.ciexyzGreen.ciexyzY, 1, fp) <= 0) {
                        return 0;
                    }
                    write_size += 4;

                    if (spFWriteLONG(&header->bV5Endpoints.ciexyzGreen.ciexyzZ, 1, fp) <= 0) {
                        return 0;
                    }
                    write_size += 4;

                    if (spFWriteLONG(&header->bV5Endpoints.ciexyzBlue.ciexyzX, 1, fp) <= 0) {
                        return 0;
                    }
                    write_size += 4;

                    if (spFWriteLONG(&header->bV5Endpoints.ciexyzBlue.ciexyzY, 1, fp) <= 0) {
                        return 0;
                    }
                    write_size += 4;

                    if (spFWriteLONG(&header->bV5Endpoints.ciexyzBlue.ciexyzZ, 1, fp) <= 0) {
                        return 0;
                    }
                    write_size += 4; /* 96 bytes */

                    if ((long)header->bV5Size > write_size) {
                        if (spFWriteDWORD(&header->bV5GammaRed, 1, fp) <= 0) {
                            return 0;
                        }
                        write_size += 4;

                        if (spFWriteDWORD(&header->bV5GammaGreen, 1, fp) <= 0) {
                            return 0;
                        }
                        write_size += 4;

                        if (spFWriteDWORD(&header->bV5GammaBlue, 1, fp) <= 0) {
                            return 0;
                        }
                        write_size += 4; /* 108 bytes */

                        if ((long)header->bV5Size > write_size) {
                            if (spFWriteDWORD(&header->bV5Intent, 1, fp) <= 0) {
                                return 0;
                            }
                            write_size += 4; /* 112 bytes */

                            if ((long)header->bV5Size > write_size) {
                                if (spFWriteDWORD(&header->bV5ProfileData, 1, fp) <= 0) {
                                    return 0;
                                }
                                write_size += 4;

                                if (spFWriteDWORD(&header->bV5ProfileSize, 1, fp) <= 0) {
                                    return 0;
                                }
                                write_size += 4; /* 120 bytes */

                                if ((long)header->bV5Size > write_size) {
                                    if (spFWriteDWORD(&header->bV5Reserved, 1, fp) <= 0) {
                                        return 0;
                                    }
                                    write_size += 4; /* 124 bytes */
                                }
                            }
                        }
                    }
                }
            }
        }
    }
	    
    if (bfOffBits > 0) {
        remain_size = (long)bfOffBits - (14 + write_size);

        if (remain_size >= 12
            && (header->bV5BitCount == 16 || header->bV5BitCount == 32)
            && (header->bV5Compression == 0 || header->bV5Compression == 3)) {
            if (spFWriteDWORD(&header->bV5RedMask, 1, fp) <= 0) {
                return 0;
            }
            write_size += 4;

            if (spFWriteDWORD(&header->bV5GreenMask, 1, fp) <= 0) {
                return 0;
            }
            write_size += 4;

            if (spFWriteDWORD(&header->bV5BlueMask, 1, fp) <= 0) {
                return 0;
            }
            write_size += 4;
            
            remain_size -= 12;
        }

        if (remain_size > 0) {
            c = 0;
            while (remain_size > 0) {
                if (fwrite(&c, 1, 1, fp) != 1) {
                    return 0;
                }
                --remain_size;
                ++write_size;
            }
        }
    }
    
    return write_size;
}

spBool spGetFileDIBInfoFromBmpV5HeaderToRead(BITMAPV5HEADER *header, spDIBitmapInfo *dibinfo)
{
    if (!(header->bV5BitCount == 16 || header->bV5BitCount == 24 || header->bV5BitCount == 32)
        || !(header->bV5Compression == 0 || header->bV5Compression == 3)) {
	spDebug(10, "spGetFileDIBInfoFromBmpV5HeaderToRead", "wrong format\n");
        return SP_FALSE;
    }

    dibinfo->width = header->bV5Width;
    if (header->bV5Height < 0) {
        dibinfo->upside_down = SP_FALSE;
        dibinfo->height = -header->bV5Height;
    } else {
        dibinfo->upside_down = SP_TRUE;
        dibinfo->height = header->bV5Height;
    }

    dibinfo->bit_size = 8;
    
    if (header->bV5AlphaMask != 0) {
        dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_BGRA;
    } else {
        if (header->bV5BitCount == 32) {
            dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_BGRN;
        } else if (header->bV5BitCount == 16) {
            dibinfo->bit_size = 5;
            dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_BGRN;
        } else {
            dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_BGR;
        }
    }
    
    return SP_TRUE;
}

spBool spSetFileDIBInfoToBmpV5HeaderToWrite(BITMAPV5HEADER *header, spDIBitmapInfo *dibinfo, unsigned long *strict_mask)
{
    long line_stride;

    if (((*strict_mask & SP_DI_BITMAP_STRICT_YUV_STATE)
	 && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_SINGLE_RGB_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_SINGLE_RGB_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_FLOAT_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_FLOAT_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK))
        || ((*strict_mask & SP_DI_BITMAP_STRICT_GRAYSCALE_STATE)
            && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK))
        || ((*strict_mask & SP_DI_BITMAP_STRICT_BIT_SIZE)
	    && !(dibinfo->bit_size == 8 || dibinfo->bit_size == 5))
	) {
	spDebug(10, "spSetPluginFileDIBInfoBmp", "wrong format\n");
	return SP_FALSE;
    }

    if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK) {
	dibinfo->pixel_format &= ~(SP_DI_PIXEL_FORMAT_YUV_MASK | SP_DI_PIXEL_FORMAT_OPTIONAL_PARAMETERS_MASK);
	dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_RGB_MASK;
    }
    if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK) {
	dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK;
	dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_ALPHA_MASK;
    }
    
    *strict_mask |= SP_DI_BITMAP_STRICT_YUV_STATE | SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE
	| SP_DI_BITMAP_STRICT_SINGLE_RGB_STATE | SP_DI_BITMAP_STRICT_FLOAT_STATE
        | SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK;
    
    memset(header, 0, sizeof(BITMAPV5HEADER));
    
    if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) {
        header->bV5Size = 108;
        header->bV5Compression = 3/* BI_BITFIELDS */;
        if (dibinfo->bit_size == 5) {
            header->bV5BitCount = 16;
            header->bV5RedMask = 0x0000f000;
            header->bV5GreenMask = 0x00000f00;
            header->bV5BlueMask = 0x000000f0;
            header->bV5AlphaMask = 0x0000000f;
        } else {
            header->bV5BitCount = 32;
            header->bV5RedMask = 0x000000ff;
            header->bV5GreenMask = 0x0000ff00;
            header->bV5BlueMask = 0x00ff0000;
            header->bV5AlphaMask = 0xff000000;
        }
    } else {
        header->bV5Size = 40;
        if (dibinfo->bit_size == 5) {
            header->bV5BitCount = 16;
            header->bV5RedMask = 0x00007c00;
            header->bV5GreenMask = 0x000003e0;
            header->bV5BlueMask = 0x0000001f;
            dibinfo->pixel_format &= ~(SP_DI_PIXEL_FORMAT_FIRST_NULL_MASK | SP_DI_PIXEL_FORMAT_LAST_NULL_MASK);
        } else {
            if ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_ORDER)
                && ((dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_LAST_NULL_MASK)
                    || (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_FIRST_NULL_MASK))) {
                header->bV5BitCount = 32;
            } else {
                dibinfo->pixel_format &= ~(SP_DI_PIXEL_FORMAT_FIRST_NULL_MASK | SP_DI_PIXEL_FORMAT_LAST_NULL_MASK);
                header->bV5BitCount = 24;
            }
        
            header->bV5RedMask = 0x00ff0000;
            header->bV5GreenMask = 0x0000ff00;
            header->bV5BlueMask = 0x000000ff;
        }
        header->bV5AlphaMask = 0;
        header->bV5Compression = 0/* BI_RGB */;
    }

    line_stride = 4 * ((dibinfo->width * header->bV5BitCount + 31) / 32);
    
    header->bV5Width = dibinfo->width;
    if (dibinfo->upside_down == SP_FALSE) {
        header->bV5Height = -dibinfo->height;
    } else {
        header->bV5Height = dibinfo->height;
    }
    header->bV5Planes = 1;
    header->bV5SizeImage = line_stride * dibinfo->height;
    header->bV5XPelsPerMeter = header->bV5YPelsPerMeter = 3780;

    /* For compatibility, value 1 which is different from specification is used. */
    /* The correct value is maybe 0x10000. */
    header->bV5GammaRed = header->bV5GammaGreen = header->bV5GammaBlue = 1; 
    
    return SP_TRUE;
}

spBool spGetDIBInfoFromBmpV5Header(BITMAPV5HEADER *header, spDIBitmapInfoInternal *dibinfo, unsigned long *strict_mask, spBool write_flag)
{
    long file_line_stride;
    
    if (((*strict_mask & SP_DI_BITMAP_STRICT_YUV_STATE)
	 && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_SINGLE_RGB_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_SINGLE_RGB_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_FLOAT_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_FLOAT_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK))
        || ((*strict_mask & SP_DI_BITMAP_STRICT_GRAYSCALE_STATE)
            && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK))
        || ((*strict_mask & SP_DI_BITMAP_STRICT_BIT_SIZE)
	    && !(dibinfo->bit_size == 8))
        || ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
            && !(dibinfo->pixel_stride == 3 || dibinfo->pixel_stride == 4))
	) {
	spDebug(10, "spGetDIBInfoFromBmpV5Header", "unsupported strict mask\n");
	return SP_FALSE;
    }
    
    dibinfo->pixel_format &= ~(SP_DI_PIXEL_FORMAT_YUV_MASK | SP_DI_PIXEL_FORMAT_OPTIONAL_PARAMETERS_MASK
			       | SP_DI_PIXEL_FORMAT_SINGLE_RGB_MASK | SP_DI_PIXEL_FORMAT_FLOAT_MASK
                               | SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK);
    
    if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK) {
	dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK;
	dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_ALPHA_MASK;
    }

    if (write_flag) {
        if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)  {
	    if ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
		&& dibinfo->pixel_stride != 4) {
		spDebug(10, "spGetDIBInfoFromBmpV5Header", "not supported pixel_stride for rgba\n");
		return SP_FALSE;
	    } else {
		dibinfo->pixel_stride = 4;
	    }
	    dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_BGRA;
	} else {
            if (dibinfo->pixel_stride == 4) {
                dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_BGRN;
            } else {
                dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_BGR;
            }
	}
	
	dibinfo->bit_size = 8;
        file_line_stride = 4 * ((dibinfo->width * (dibinfo->pixel_stride * 8) + 31) / 32);

	if ((*strict_mask & SP_DI_BITMAP_STRICT_LINE_STRIDE)
	    && dibinfo->line_stride != file_line_stride) {
	} else {
	    dibinfo->line_stride = file_line_stride;
	}
	
	if (!(*strict_mask & SP_DI_BITMAP_STRICT_UPSIDE_DOWN_STATE)) {
	    dibinfo->upside_down = SP_TRUE;
	}
    } else {
        dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_RGB_MASK;

	if ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_ORDER)
	    && !(dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK)) {
	} else {
	    dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_INVERSE_MASK;
	}

	if (header->bV5AlphaMask != 0) {
	    spDebug(80, "spGetDIBInfoFromBmpV5Header", "alpha channel included\n");
	
	    if (((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
		 && dibinfo->pixel_stride == 3)
		|| ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_STATE)
		    && !(dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)))  {
                return SP_FALSE;
	    } else {
		dibinfo->pixel_stride = 4;
		if ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_ORDER)
		    && !(dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_LAST_ALPHA_MASK)) {
		    dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK | SP_DI_PIXEL_FORMAT_ALPHA_MASK;
		} else {
		    dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK;
		    dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_LAST_ALPHA_MASK | SP_DI_PIXEL_FORMAT_ALPHA_MASK;
		}
	    }
	} else {
	    spDebug(80, "spGetDIBInfoFromBmpV5Header", "no alpha channel\n");
	
	    dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_ALPHA_MASK;
	
            spDebug(80, "spGetDIBInfoFromBmpV5Header", "pixel_stride = %ld, vB5BitCount = %ld\n",
                    dibinfo->pixel_stride, header->bV5BitCount);
            
            if ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
                && dibinfo->pixel_stride == 4) {
                if (header->bV5BitCount != 32) {
                    return SP_FALSE;
                }
            } else if ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
                       && dibinfo->pixel_stride == 3) {
                if (header->bV5BitCount > 24) {
                    return SP_FALSE;
                }
	    } else {
		dibinfo->pixel_stride = header->bV5BitCount / 8;
	    }
            
            if (header->bV5BitCount > 24) {
                if (!(*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_ORDER)) {
                    dibinfo->pixel_format &= ~SP_DI_PIXEL_FORMAT_FIRST_NULL_MASK;
                    dibinfo->pixel_format |= SP_DI_PIXEL_FORMAT_LAST_ALPHA_MASK;
                }
            }
	}

	dibinfo->width = header->bV5Width;
	dibinfo->height = ABS(header->bV5Height);

	dibinfo->bit_size = 8;
	dibinfo->line_stride = 4 * ((dibinfo->width * (dibinfo->pixel_stride * 8) + 31) / 32);
	if (!(*strict_mask & SP_DI_BITMAP_STRICT_UPSIDE_DOWN_STATE)) {
	    dibinfo->upside_down = SP_FALSE;
	}
	spDebug(80, "spGetDIBInfoFromBmpV5Header", "pixel_format = %lx, bit_size = %ld, pixel_stride = %ld, line_stride = %ld\n",
		dibinfo->pixel_format, dibinfo->bit_size, dibinfo->pixel_stride, dibinfo->line_stride);

	*strict_mask |= SP_DI_BITMAP_STRICT_YUV_STATE | SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE
	    | SP_DI_BITMAP_STRICT_SINGLE_RGB_STATE | SP_DI_BITMAP_STRICT_FLOAT_STATE
            | SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK;
    }
    
    spDebug(80, "spGetDIBInfoFromBmpV5Header", "done\n");
    
    return SP_TRUE;
}

static long readBmpV5DataWithConversion(BITMAPV5HEADER *header, spDIPixelFormat pixel_format,
                                        long line_stride, long file_line_stride, spBool vert_flip,
                                        unsigned char *data, FILE *fp)
{
    long x, y;
    int i, j;
    int nmask;
    long height;
    spLong32 mask;
    DWORD masks[4];
    DWORD divs[4];
    double maskweights[4];
    DWORD pixel;
    WORD spixel;
    unsigned char c;
    unsigned char rgba[4] = {0, 0, 0, 0};
    double d;
    long total_nread, line_nread;
    long skip_length;
    unsigned char *ptr;
    
    masks[0] = header->bV5RedMask;
    masks[1] = header->bV5GreenMask;
    masks[2] = header->bV5BlueMask;
    masks[3] = header->bV5AlphaMask;

    if (header->bV5BitCount == 32) {
        nmask = 4;
    } else {
        nmask = 3;
    }
    spDebug(50, "readBmpV5DataWithConversion", "pixel_format = %lx, bV5BitCount = %ld, nmask = %d\n",
            pixel_format, header->bV5BitCount, nmask);

    if (masks[0] == 0 && masks[1] == 0 && masks[2] == 0) {
        if (header->bV5BitCount == 16) {
            if (pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) {
                masks[0] = 0x0000f000;
                masks[1] = 0x00000f00;
                masks[2] = 0x000000f0;
                masks[3] = 0x0000000f;
            } else {
                masks[0] = 0x00007c00;
                masks[1] = 0x000003e0;
                masks[2] = 0x0000001f;
                masks[3] = 0;
            }
        } else if (header->bV5BitCount == 32 && pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) {
            masks[0] = 0x000000ff;
            masks[1] = 0x0000ff00;
            masks[2] = 0x00ff0000;
            masks[3] = 0xff000000;
        } else {
            masks[0] = 0x00ff0000;
            masks[1] = 0x0000ff00;
            masks[2] = 0x000000ff;
            masks[3] = 0;
        }
    }

    spDebug(50, "readBmpV5DataWithConversion", "masks = %lx, %lx, %lx, %lx\n",
            masks[0], masks[1], masks[2], masks[3]);
    
    for (i = 0; i < nmask; i++) {
        mask = (spLong32)masks[i];
        divs[i] = (DWORD)(mask & -mask);
        if (divs[i] == 0) divs[i] = 1;

        if (mask == 0) {
            maskweights[i] = 1.0;
        } else {
            maskweights[i] = (double)(mask / divs[i]);
        }
    }

    height = ABS(header->bV5Height);
    if (vert_flip == SP_TRUE) {
        ptr = data + line_stride * (height - 1);
    } else {
        ptr = data;
    }
    total_nread = 0;
    for (y = 0; y < height; y++) {
        line_nread = 0;
        for (x = 0; x < header->bV5Width; x++) {
            if (header->bV5BitCount == 16) {
                if (spFReadWORD(&spixel, 1, fp) <= 0) {
                    spDebug(50, "readBmpV5DataWithConversion", "fread failed: total_nread = %ld\n", total_nread);
                    return 0;
                }
                pixel = spixel;
                line_nread += 2;
            } else {
                if (spFReadDWORD(&pixel, 1, fp) <= 0) {
                    spDebug(50, "readBmpV5DataWithConversion", "fread failed: total_nread = %ld\n", total_nread);
                    return 0;
                }
                line_nread += 4;
            }
            
            if (header->bV5BitCount == 16) {
                for (i = 0; i < nmask; i++) {
                    d = (double)((pixel & masks[i]) / divs[i]);
                    c = (unsigned char)spRound((d * 255.0) / maskweights[i]);
                    if (pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK) {
                        ptr[nmask * x + nmask - i - 1] = c;
                    } else {
                        ptr[nmask * x + i] = c;
                    }
                }
            } else if (header->bV5BitCount == 32) {
                for (i = 0; i < nmask; i++) {
                    if (maskweights[i] != 1.0) {
                        rgba[i] = (unsigned char)spRound(255.0 * (double)((pixel & masks[i]) / divs[i]) / maskweights[i]);
                    } else {
                        rgba[i] = (unsigned char)((pixel & masks[i]) / divs[i]);
                    }
                }
                
                if (pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK) {
                    ptr[nmask * x] = rgba[3];
                    j = 1;
                } else {
                    ptr[nmask * x + 3] = rgba[3];
                    j = 0;
                }
                for (i = 0; i < 3; i++) {
                    if (pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK) {
                        ptr[nmask * x + j + i] = rgba[2 - i];
                    } else {
                        ptr[nmask * x + j + i] = rgba[i];
                    }
                }
            } else {
                for (i = 0; i < nmask; i++) {
                    if (maskweights[i] != 1.0) {
                        c = (unsigned char)spRound(255.0 * (double)((pixel & masks[i]) / divs[i]) / maskweights[i]);
                    } else {
                        c = (unsigned char)((pixel & masks[i]) / divs[i]);
                    }
                    if (pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK) {
                        ptr[nmask * x + nmask - i - 1] = c;
                    } else {
                        ptr[nmask * x + i] = c;
                    }
                }
            }
        }

        skip_length = file_line_stride - line_nread;

        if (skip_length > 0) {
            if (fseek(fp, skip_length, SEEK_CUR) != 0) {
                return 0;
            }
            line_nread += skip_length;
        }
        
        total_nread += line_nread;
        if (vert_flip == SP_TRUE) {
            ptr -= line_stride;
        } else {
            ptr += line_stride;
        }
    }

    return total_nread;
}

/* Pixel size conversion is not supported such as: BGR file --> RGBN */
long spReadBmpV5Data(BITMAPV5HEADER *header, spDIPixelFormat pixel_format, long line_stride,
                     spBool upside_down, unsigned char *data, FILE *fp)
{
    long x, y;
    long height;
    long nread, total_nread;
    long file_line_stride;
    long skip_length;
    unsigned char r, g, b, a;
    unsigned char *ptr;
    spBool vert_flip;
    spBool rgba_same_order = SP_FALSE;

    file_line_stride = 4 * ((header->bV5Width * header->bV5BitCount + 31) / 32);
    if (line_stride <= 0) line_stride = file_line_stride;

    spDebug(50, "spReadBmpV50Data", "line_stride = %ld, file_line_stride = %ld\n",
            line_stride, file_line_stride);
    
    if (header->bV5Compression == 3 && header->bV5BitCount == 32 && header->bV5AlphaMask != 0) {
        if (pixel_format == SP_DI_PIXEL_FORMAT_BGRA) {
            if (header->bV5RedMask == 0x00ff0000
                && header->bV5GreenMask == 0x0000ff00
                && header->bV5BlueMask == 0x000000ff
                && header->bV5AlphaMask == 0xff000000) {
                rgba_same_order = SP_TRUE;
            }
        } else if (pixel_format == SP_DI_PIXEL_FORMAT_RGBA) {
            if (header->bV5RedMask == 0x000000ff
                && header->bV5GreenMask == 0x0000ff00
                && header->bV5BlueMask == 0x00ff0000
                && header->bV5AlphaMask == 0xff000000) {
                rgba_same_order = SP_TRUE;
            }
        }
    }

    if ((upside_down == SP_FALSE && header->bV5Height > 0)
        || (upside_down == SP_TRUE && header->bV5Height < 0)) {
        vert_flip = SP_TRUE;
    } else {
        vert_flip = SP_FALSE;
    }

    if (line_stride == file_line_stride && vert_flip == SP_FALSE
        && ((header->bV5Compression == 0 && header->bV5BitCount == 24 && pixel_format == SP_DI_PIXEL_FORMAT_BGR)
            || (header->bV5Compression == 0 && header->bV5BitCount == 32 && pixel_format == SP_DI_PIXEL_FORMAT_BGRN)
            || rgba_same_order == SP_TRUE)) {
        total_nread = (long)fread(data, 1, line_stride * header->bV5Width, fp);
    } else if (header->bV5Compression == 0 && (header->bV5BitCount == 24 || header->bV5BitCount == 32)) {
        height = ABS(header->bV5Height);
            
        if (vert_flip == SP_TRUE) {
            ptr = data + line_stride * (height - 1);
        } else {
            ptr = data;
        }
        total_nread = 0;
        for (y = 0; y < height; y++) {
            if ((nread = (long)fread(ptr, 1, line_stride, fp)) <= 0) {
                return 0;
            }

            if (header->bV5BitCount == 24 && pixel_format == SP_DI_PIXEL_FORMAT_RGB) {
                for (x = 0; x < header->bV5Width; x++) {
                    b = ptr[3 * x];
                    r = ptr[3 * x + 2];
                    ptr[3 * x] = r;
                    ptr[3 * x + 2] = b;
                }
            } else if (header->bV5BitCount == 32 && pixel_format == SP_DI_PIXEL_FORMAT_RGBN) {
                for (x = 0; x < header->bV5Width; x++) {
                    b = ptr[4 * x];
                    r = ptr[4 * x + 2];
                    ptr[4 * x] = r;
                    ptr[4 * x + 2] = b;
                }
            } else if (header->bV5BitCount == 32 && pixel_format == SP_DI_PIXEL_FORMAT_NRGB) {
                for (x = 0; x < header->bV5Width; x++) {
                    b = ptr[4 * x];
                    g = ptr[4 * x + 1];
                    r = ptr[4 * x + 2];
                    a = ptr[4 * x + 3];
                    ptr[4 * x] = a;
                    ptr[4 * x + 1] = r;
                    ptr[4 * x + 2] = g;
                    ptr[4 * x + 3] = b;
                }
            } else if (header->bV5BitCount == 32 && pixel_format == SP_DI_PIXEL_FORMAT_NBGR) {
                for (x = 0; x < header->bV5Width; x++) {
                    b = ptr[4 * x];
                    g = ptr[4 * x + 1];
                    r = ptr[4 * x + 2];
                    a = ptr[4 * x + 3];
                    ptr[4 * x] = a;
                    ptr[4 * x + 1] = b;
                    ptr[4 * x + 2] = g;
                    ptr[4 * x + 3] = r;
                }
            }

            skip_length = file_line_stride - nread;

            if (skip_length > 0) {
                if (fseek(fp, skip_length, SEEK_CUR) != 0) {
                    return 0;
                }
            }

            total_nread += file_line_stride;
            
            if (vert_flip == SP_TRUE) {
                ptr -= line_stride;
            } else {
                ptr += line_stride;
            }
        }
    } else {
        total_nread = readBmpV5DataWithConversion(header, pixel_format, line_stride, file_line_stride, vert_flip, data, fp);
    }
    
    return total_nread;
}

#if 1
static long writeBmpV5DataWithConversion(BITMAPV5HEADER *header, spDIPixelFormat pixel_format,
                                         long line_stride, long file_line_stride, spBool vert_flip,
                                         unsigned char *data, FILE *fp)
{
    long x, y;
    int i, j;
    int nmask;
    long height;
    spLong32 mask;
    DWORD masks[4];
    DWORD divs[4];
    double maskweights[4];
    DWORD pixel;
    WORD spixel;
    unsigned char c;
    unsigned char rgba[4];
    double d;
    long total_nwrite, line_nwrite;
    long skip_length;
    unsigned char *ptr;
    
    masks[0] = header->bV5RedMask;
    masks[1] = header->bV5GreenMask;
    masks[2] = header->bV5BlueMask;
    masks[3] = header->bV5AlphaMask;

    if (header->bV5BitCount == 32) {
        nmask = 4;
    } else {
        nmask = 3;
    }
    spDebug(50, "writeBmpV5DataWithConversion", "pixel_format = %lx, bV5BitCount = %ld, nmask = %d\n",
            pixel_format, header->bV5BitCount, nmask);

    if (masks[0] == 0 && masks[1] == 0 && masks[2] == 0) {
        if (header->bV5BitCount == 16) {
            masks[0] = 0x00007c00;
            masks[1] = 0x000003e0;
            masks[2] = 0x0000001f;
            masks[3] = 0;
        } else if (pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) {
            masks[0] = 0x000000ff;
            masks[1] = 0x0000ff00;
            masks[2] = 0x00ff0000;
            masks[3] = 0xff000000;
        } else {
            masks[0] = 0x00ff0000;
            masks[1] = 0x0000ff00;
            masks[2] = 0x000000ff;
            masks[3] = 0;
        }
    }

    for (i = 0; i < nmask; i++) {
        mask = (spLong32)masks[i];
        divs[i] = (DWORD)(mask & -mask);
        if (divs[i] == 0) divs[i] = 1;

        if (mask == 0) {
            maskweights[i] = 1.0;
        } else {
            maskweights[i] = (double)(mask / divs[i]);
        }
    }

    height = ABS(header->bV5Height);
    if (vert_flip == SP_TRUE) {
        ptr = data + line_stride * (height - 1);
    } else {
        ptr = data;
    }
    total_nwrite = 0;
    for (y = 0; y < height; y++) {
        line_nwrite = 0;
        for (x = 0; x < header->bV5Width; x++) {
            pixel = 0;
            if (header->bV5BitCount == 16) {
                for (i = 0; i < nmask; i++) {
                    if (pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK) {
                        c = ptr[nmask * x + nmask - i - 1];
                    } else {
                        c = ptr[nmask * x + i];
                    }
                    d = spRound(maskweights[i] * (double)c / 255.0);
                    pixel |= (masks[i] & ((DWORD)d * divs[i]));
                }
            } else if (header->bV5BitCount == 32) {
                if (pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK) {
                    rgba[3] = ptr[nmask * x];
                    j = 1;
                } else {
                    rgba[3] = ptr[nmask * x + 3];
                    j = 0;
                }
                for (i = 0; i < 3; i++) {
                    if (pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK) {
                        rgba[2 - i] = ptr[nmask * x + j + i];
                    } else {
                        rgba[i]= ptr[nmask * x + j + i];
                    }
                }
                
                for (i = 0; i < nmask; i++) {
                    if (maskweights[i] != 1.0) {
                        d = spRound(maskweights[i] * (double)rgba[i] / 255.0);
                        pixel |= (masks[i] & ((DWORD)d * divs[i]));
                    } else {
                        pixel |= (masks[i] & ((DWORD)rgba[i] * divs[i]));
                    }
                }
            } else {
                for (i = 0; i < nmask; i++) {
                    if (pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK) {
                        c = ptr[nmask * x + nmask - i - 1];
                    } else {
                        c = ptr[nmask * x + i];
                    }
                    if (maskweights[i] != 1.0) {
                        d = spRound(maskweights[i] * (double)c / 255.0);
                        pixel |= (masks[i] & ((DWORD)d * divs[i]));
                    } else {
                        pixel |= (masks[i] & ((DWORD)c * divs[i]));
                    }
                }
            }
            
            if (header->bV5BitCount == 16) {
                spixel = (WORD)pixel;
                if (spFWriteWORD(&spixel, 1, fp) <= 0) {
                    return 0;
                }
                line_nwrite += 2;
            } else {
                if (spFWriteDWORD(&pixel, 1, fp) <= 0) {
                    return 0;
                }
                line_nwrite += 4;
            }
        }

        skip_length = file_line_stride - line_nwrite;

        if (skip_length > 0) {
            x = skip_length;
            c = 0;
            while (x > 0) {
                if (fwrite(&c, 1, 1, fp) != 1) {
                    return 0;
                }
                --x;
                ++line_nwrite;
            }
        }

        total_nwrite += line_nwrite;
        if (vert_flip == SP_TRUE) {
            ptr -= line_stride;
        } else {
            ptr += line_stride;
        }
    }

    return total_nwrite;
}

long spWriteBmpV5Data(BITMAPV5HEADER *header, spDIPixelFormat pixel_format, long line_stride,
                      spBool upside_down, unsigned char *data, FILE *fp)
{
    long x, y;
    long height;
    long nwrite, total_nwrite, line_nwrite;
    long file_line_stride;
    long skip_length;
    long pixel_stride, file_pixel_stride;
    unsigned char pixel[4];
    unsigned char *ptr;
    spBool vert_flip;
    spBool rgba_same_order = SP_FALSE;

    file_line_stride = 4 * ((header->bV5Width * header->bV5BitCount + 31) / 32);
    if (line_stride <= 0) line_stride = file_line_stride;

    spDebug(50, "spWriteBmpV5Data", "line_stride = %ld, file_line_stride = %ld\n",
            line_stride, file_line_stride);
    
    if (header->bV5Compression == 3 && header->bV5BitCount == 32 && header->bV5AlphaMask != 0) {
        if (pixel_format == SP_DI_PIXEL_FORMAT_BGRA) {
            if (header->bV5RedMask == 0x00ff0000
                && header->bV5GreenMask == 0x0000ff00
                && header->bV5BlueMask == 0x000000ff
                && header->bV5AlphaMask == 0xff000000) {
                rgba_same_order = SP_TRUE;
            }
        } else if (pixel_format == SP_DI_PIXEL_FORMAT_RGBA) {
            if (header->bV5RedMask == 0x000000ff
                && header->bV5GreenMask == 0x0000ff00
                && header->bV5BlueMask == 0x00ff0000
                && header->bV5AlphaMask == 0xff000000) {
                rgba_same_order = SP_TRUE;
            }
        }
    }

    if ((upside_down == SP_FALSE && header->bV5Height > 0)
        || (upside_down == SP_TRUE && header->bV5Height < 0)) {
        vert_flip = SP_TRUE;
    } else {
        vert_flip = SP_FALSE;
    }

    if (line_stride == file_line_stride && vert_flip == SP_FALSE
        && ((header->bV5Compression == 0 && header->bV5BitCount == 24 && pixel_format == SP_DI_PIXEL_FORMAT_BGR)
            || (header->bV5Compression == 0 && header->bV5BitCount == 32 && pixel_format == SP_DI_PIXEL_FORMAT_BGRN)
            || rgba_same_order == SP_TRUE)) {
        total_nwrite = (long)fwrite(data, 1, line_stride * header->bV5Width, fp);
    } else if (header->bV5Compression == 0 && (header->bV5BitCount == 24 || header->bV5BitCount == 32)) {
        height = ABS(header->bV5Height);

        if ((pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK) || (pixel_format & SP_DI_PIXEL_FORMAT_LAST_ALPHA_MASK)) {
            pixel_stride = 4;
        } else {
            pixel_stride = 3;
        }
        file_pixel_stride = header->bV5BitCount / 8;
            
        if (vert_flip == SP_TRUE) {
            ptr = data + line_stride * (height - 1);
        } else {
            ptr = data;
        }
        total_nwrite = 0;
        for (y = 0; y < height; y++) {
            for (x = 0, line_nwrite = 0; x < header->bV5Width; x++) {
                pixel[3] = 0;
                
                if (header->bV5BitCount == 24 && pixel_format == SP_DI_PIXEL_FORMAT_RGB) {
                    pixel[0] = ptr[3 * x + 2];
                    pixel[1] = ptr[3 * x + 1];
                    pixel[2] = ptr[3 * x];
                } else if (header->bV5BitCount == 32 && pixel_format == SP_DI_PIXEL_FORMAT_RGBN) {
                    pixel[0] = ptr[4 * x + 2];
                    pixel[1] = ptr[4 * x + 1];
                    pixel[2] = ptr[4 * x];
                    pixel[3] = ptr[4 * x + 3];
                } else if (header->bV5BitCount == 32 && pixel_format == SP_DI_PIXEL_FORMAT_NRGB) {
                    pixel[0] = ptr[4 * x + 3];
                    pixel[1] = ptr[4 * x + 2];
                    pixel[2] = ptr[4 * x + 1];
                    pixel[3] = ptr[4 * x];
                } else if (header->bV5BitCount == 32 && pixel_format == SP_DI_PIXEL_FORMAT_NBGR) {
                    pixel[0] = ptr[4 * x + 1];
                    pixel[1] = ptr[4 * x + 2];
                    pixel[2] = ptr[4 * x + 3];
                    pixel[3] = ptr[4 * x];
                } else {
                    memcpy(pixel, &ptr[4 * x], pixel_stride);
                }

                if ((nwrite = (long)fwrite(pixel, 1, file_pixel_stride, fp)) <= 0) {
                    return 0;
                }
                line_nwrite += nwrite;
            }

            skip_length = file_line_stride - line_nwrite;

            if (skip_length > 0) {
                x = skip_length;
                pixel[0] = 0;
                while (x > 0) {
                    if (fwrite(pixel, 1, 1, fp) != 1) {
                        return 0;
                    }
                    --x;
                }
            }

            total_nwrite += file_line_stride;
            
            if (vert_flip == SP_TRUE) {
                ptr -= line_stride;
            } else {
                ptr += line_stride;
            }
        }
    } else {
        total_nwrite = writeBmpV5DataWithConversion(header, pixel_format, line_stride, file_line_stride, vert_flip, data, fp);
    }
    
    return total_nwrite;
}
#endif
