#include <sp/spBaseLib.h>
#include <sp/spComponentLib.h>
#include <sp/spCanvasP.h>
#include <sp/spGraphicsP.h>
#include <sp/spDrawP.h>

#include <sp/spDIBitmapP.h>

extern spTopLevel sp_toplevel;

int spGetDIBitmapDefaultPixelSizeXm(unsigned int *depth)
{
    int bitmap_pad;
    
    *depth = DefaultDepthOfScreen(SpTopLevelArch(sp_toplevel).screen);
#if 1
    if (*depth <= 16) {
	bitmap_pad = 16;
    } else {
	bitmap_pad = 32;
    }
#else
    bitmap_pad = *depth;
#endif

    return bitmap_pad;
}

spBool spCheckBasicDIBitmapStrictMaskXm(spDIBitmapInfoInternal *info, unsigned long strict_mask,
					unsigned int *pdepth, int *pbitmap_pad, long *pline_stride)
{
    unsigned int depth;
    int bitmap_pad;
    long line_stride;

    if (((strict_mask & SP_DI_BITMAP_STRICT_GRAYSCALE_STATE)
	 && (info->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_YUV_STATE)
	    && (info->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_ALPHA_STATE)
	    && (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_PIXEL_ORDER)
	    && (info->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_UPSIDE_DOWN_STATE)
	    && info->upside_down != SP_FALSE)) {
	return SP_FALSE;
    }
    
    bitmap_pad = spGetDIBitmapDefaultPixelSizeXm(&depth);
    
    if (strict_mask & SP_DI_BITMAP_STRICT_BIT_SIZE) {
	if (info->bit_size * 3 != depth) {
	    spDebug(10, "spCheckBasicDIBitmapStrictMaskXm", "bit_size %ld is wrong\n", info->bit_size);
	    return SP_FALSE;
	}
    }
    
    if (strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE) {
	if (info->pixel_stride * 8 != bitmap_pad) {
	    spDebug(10, "spCheckBasicDIBitmapStrictMaskXm", "pixel_stride %ld is wrong\n", info->pixel_stride);
	    return SP_FALSE;
	}
    }
    
    line_stride = (bitmap_pad / 8) * info->width;
    if (strict_mask & SP_DI_BITMAP_STRICT_LINE_STRIDE) {
	if (line_stride != info->line_stride) {
	    spDebug(10, "spCheckBasicDIBitmapStrictMaskXm", "line_stride %ld is wrong\n", info->line_stride);
	    return SP_FALSE;
	}
    }

    if (pdepth != NULL) *pdepth = depth;
    if (pbitmap_pad != NULL) *pbitmap_pad = bitmap_pad;
    if (pline_stride != NULL) *pline_stride = line_stride;

    return SP_TRUE;
}


spBool spUpdateDIBitmapInfoArch(spDIBitmapInfoInternal *info, unsigned long *strict_mask)
{
    unsigned int depth;
    int bitmap_pad;
    long line_stride;

    if (spCheckBasicDIBitmapStrictMaskXm(info, *strict_mask, &depth, &bitmap_pad, &line_stride) == SP_FALSE) {
	return SP_FALSE;
    }
    
    info->upside_down = SP_FALSE;
    info->bit_size = depth / 3;
    info->pixel_stride = bitmap_pad / 8;
    info->line_stride = line_stride;

    if (info->pixel_stride == 4) {
	info->pixel_format = SP_DI_PIXEL_FORMAT_RGBN;
    } else {
	info->pixel_format = SP_DI_PIXEL_FORMAT_RGB;
    }

    *strict_mask |= SP_DI_BITMAP_STRICT_ALL;
    
    return SP_TRUE;
}

void *spCreateDIBitmapArch(spDIBitmapInfoInternal *info, unsigned long strict_mask)
{
    XImage *ximage = NULL;
    unsigned int depth;
    int bitmap_pad;
    long line_stride;
    char *buf;
    long buf_size;

    if (spCheckBasicDIBitmapStrictMaskXm(info, strict_mask, &depth, &bitmap_pad, &line_stride) == SP_FALSE) {
	return NULL;
    }
    
    buf_size = line_stride * info->height;
    spDebug(80, "spCreateDIBitmapArch", "buf_size = %ld\n", buf_size);
    if ((buf = malloc(buf_size)) != NULL) {
	ximage = XCreateImage(SpTopLevelArch(sp_toplevel).display,
			      DefaultVisualOfScreen(SpTopLevelArch(sp_toplevel).screen),
			      depth, ZPixmap, 0, buf,
			      info->width, info->height, bitmap_pad, 0);
	if (ximage != NULL) {
	    info->upside_down = SP_FALSE;
	    info->bit_size = depth / 3;
	    info->pixel_stride = bitmap_pad / 8;
	    info->line_stride = line_stride;
	    if (info->pixel_stride == 4) {
		info->pixel_format = SP_DI_PIXEL_FORMAT_RGBN;
	    } else {
		info->pixel_format = SP_DI_PIXEL_FORMAT_RGB;
	    }
	    info->data = (unsigned char *)ximage->data;
	}
    }

    spDebug(80, "spCreateDIBitmapArch", "done: ximage = %ld\n", (long)ximage);
    
    return ximage;
}

spBool spDestroyDIBitmapArch(void *bitmap)
{
    XDestroyImage((XImage *)bitmap);
    return SP_TRUE;
}


spBool spSetDIBDataRGBAArch(spDIBitmapInfoInternal *info, void *dibdata, int x, int y, int r, int g, int b, int a)
{
    unsigned long pixel;
	
    pixel = (((unsigned long)r) << 16) | (((unsigned long)g) << 8) | (((unsigned long)b));
	
    XPutPixel((XImage *)dibdata, x, y, pixel);

    return SP_TRUE;
}

spBool spAddDIBDataRGBAArch(spDIBitmapInfoInternal *info, void *dibdata, int x, int y, int r, int g, int b, int a)
{
    int r0, g0, b0;
    unsigned long pixel;

    pixel = XGetPixel((XImage *)dibdata, x, y);
	
    r0 = (int)(0xFF & (pixel >> 16));
    g0 = (int)(0xFF & (pixel >> 8));
    b0 = (int)(0xFF & (pixel));
    r0 += r;
    g0 += g;
    b0 += b;
    r0 = MIN(MAX(r0, 0), 255);
    g0 = MIN(MAX(g0, 0), 255);
    b0 = MIN(MAX(b0, 0), 255);
    pixel = (((unsigned long)r0) << 16) | (((unsigned long)g0) << 8) | (((unsigned long)b0));
	
    XPutPixel((XImage *)dibdata, x, y, pixel);

    return SP_TRUE;
}

spBool spGetDIBDataRGBAArch(spDIBitmapInfoInternal *info, void *dibdata, int x, int y, int *r, int *g, int *b, int *a)
{
    unsigned long pixel;

    pixel = XGetPixel((XImage *)dibdata, x, y);
	
    *r = (int)(0xFF & (pixel >> 16));
    *g = (int)(0xFF & (pixel >> 8));
    *b = (int)(0xFF & (pixel));
    if (a != NULL) *a = 255;

    return SP_TRUE;
}

spBool spCopyDIBitmapToImageArch(spDIBitmapInfoInternal *info, void *bitmap, spComponent dest, int src_x, int src_y,
				 int width, int height, int dest_x, int dest_y)
{
    XPutImage(SpTopLevelArch(sp_toplevel).display,
	      SpPrimitiveArch(dest).pixmap,
	      SpTopLevelArch(sp_toplevel).fg_gc,
	      (XImage *)bitmap,
	      src_x, src_y, dest_x, dest_y,
	      MIN(info->width - src_x, width),
	      MIN(info->height - src_y, height));

    return SP_TRUE;
}

void *spCopyImageToDIBitmapArch(spComponent src, spDIBitmapInfoInternal *info, void *bitmap, int src_x, int src_y,
				int width, int height, int dest_x, int dest_y)
{
    XImage *ximage = (XImage *)bitmap;
    
    XGetSubImage(SpTopLevelArch(sp_toplevel).display,
		 SpPrimitiveArch(src).pixmap,
		 src_x, src_y,
		 MIN(SpComponentPart(src).client_width - src_x, width),
		 MIN(SpComponentPart(src).client_height - src_y, height),
		 AllPlanes, ZPixmap,
		 ximage, dest_x, dest_y);

    return bitmap;
}

typedef struct _spDIBitmapInstanceArch {
    char *filename;
    Pixmap pixmap;
    int width;
    int height;
} *spDIBitmapInstanceArch;

void *spInitDIBitmapInstanceArch(const char *name)
{
    spDIBitmapInstanceArch dibinstance = NULL;
    
    dibinstance = xalloc(1, struct _spDIBitmapInstanceArch);
    memset(dibinstance, 0, sizeof(struct _spDIBitmapInstanceArch));
    
    spDebug(80, "spInitDIBitmapInstanceArch", "done\n");
    
    return dibinstance;
}

spBool spFreeDIBitmapInstanceArch(void *instance)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    xfree(dibinstance);
    
    spDebug(80, "spFreeDIBitmapInstanceArch", "done\n");
    
    return SP_TRUE;
}

int spGetDIBitmapPriorityArch(void *instance)
{
    return SP_PLUGIN_PRIORITY_HIGHER;
}

unsigned long spGetDIBitmapCapsArch(void *instance)
{
    return SP_DI_BITMAP_CAPS_READ | SP_DI_BITMAP_CAPS_WRITE
	| SP_DI_BITMAP_CAPS_READ_USE_BITMAP_ARGUMENT | SP_DI_BITMAP_CAPS_WRITE_USE_BITMAP_ARGUMENT
	| SP_DI_BITMAP_CAPS_PIXEL_STRIDE_3 | SP_DI_BITMAP_CAPS_COPY_IMAGE | SP_DI_BITMAP_CAPS_COPY_IMAGE_STRETCHABLE
	| SP_DI_BITMAP_CAPS_UPSIDE_UP | SP_DI_BITMAP_CAPS_NONINVERSE_FORMAT;
}

static const char *sp_dibitmap_file_type_list_arch[] = {
    "xpm",
    NULL,
};

static const char *sp_dibitmap_file_desc_list_arch[] = {
    "XPM",
    NULL,
};

static const char *sp_dibitmap_file_filter_list_arch[] = {
    "*.xpm",
    NULL,
};

const char **spGetDIBitmapFileTypeListArch(void *instance)
{
    return sp_dibitmap_file_type_list_arch;
}

const char **spGetDIBitmapFileDescListArch(void *instance)
{
    return sp_dibitmap_file_desc_list_arch;
}

const char **spGetDIBitmapFileFilterListArch(void *instance)
{
    return sp_dibitmap_file_filter_list_arch;
}

spBool spIsDIBitmapSupportedArch(void *instance, const char *filename)
{
    spBool supported = SP_FALSE;

    spDebug(80, "spIsDIBitmapSupportedArch", "in: filename = %s\n", filename);
    
    if (spGetBaseName(filename) == filename && filename[0] == '.') {
	if (strcaseeq(filename, ".xpm")) {
	    supported = SP_TRUE;
	}
    } else {
	Pixmap pixmap;
	
	if ((pixmap = spCreatePixmapFromDataOrFileXm(NULL, NULL, filename, NULL, NULL)) != None) {
	    supported = SP_TRUE;
	    XFreePixmap(SpTopLevelArch(sp_toplevel).display, pixmap);
	} else {
	    spDebug(80, "spIsDIBitmapSupportedArch", "spCreatePixmapFromDataOrFileXm failed\n");
	}
    }
    
    spDebug(80, "spIsDIBitmapSupportedArch", "done: supported = %d\n", supported);
    
    return supported;
}

spBool spSetDIBitmapFileTypeArch(void *instance, int index)
{
    if (index == 0) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

int spGetDIBitmapFileTypeArch(void *instance)
{
    return 0;
}

const char *spGetDIBitmapBestSuffixArch(void *instance)
{
    return ".xpm";
}

spBool spOpenDIBitmapFileArch(void *instance, const char *filename, const char *mode)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    if (mode[0] == 'w') {
	dibinstance->width = dibinstance->height = 0;
	dibinstance->filename = strclone(filename);
	
	return SP_TRUE;
    } else {
	dibinstance->filename = NULL;
	dibinstance->pixmap = spCreatePixmapFromDataOrFileXm(NULL, NULL, filename,
							     &dibinstance->width, &dibinstance->height);
	if (dibinstance->pixmap != None) {
	    return SP_TRUE;
	} else {
	    return SP_FALSE;
	}
    }
}

spBool spCloseDIBitmapFileArch(void *instance)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    if (dibinstance->pixmap != None) {
	XFreePixmap(SpTopLevelArch(sp_toplevel).display, dibinstance->pixmap);
	dibinstance->pixmap = None;
    }
    if (dibinstance->filename != NULL) {
	xfree(dibinstance->filename);
	dibinstance->filename = NULL;
    }
    
    return SP_TRUE;
}

spBool spGetDIBitmapInfoArch(void *instance, spDIBitmapInfoInternal *info, unsigned long *strict_mask)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    if (dibinstance->pixmap != None) {
        unsigned int depth;
	int bitmap_pad;
	long line_stride;

	if (spCheckBasicDIBitmapStrictMaskXm(info, *strict_mask, &depth, &bitmap_pad, &line_stride) == SP_FALSE) {
	    return SP_FALSE;
	}
	
	info->width = dibinstance->width;
	info->height = dibinstance->height;
    
	info->pixel_format = SP_DI_PIXEL_FORMAT_RGB;
	info->upside_down = SP_FALSE;
	info->bit_size = depth / 3;
	info->pixel_stride = bitmap_pad / 8;
	info->line_stride = line_stride;

	*strict_mask |= SP_DI_BITMAP_STRICT_ALL;
	
	return SP_TRUE;
    } 
    
    return SP_FALSE;
}

void *spReadDIBitmapArch(void *instance, spDIBitmapInfoInternal *info, void *bitmap)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    XImage *ximage = (XImage *)bitmap;

    if (dibinstance->pixmap != None) {
	XGetSubImage(SpTopLevelArch(sp_toplevel).display, dibinstance->pixmap,
		     0, 0,
		     info->width, info->height,
		     AllPlanes, ZPixmap,
		     ximage, 0, 0);
	
	return ximage;
    } else {
	return NULL;
    }
}

spBool spWriteDIBitmapArch(void *instance, spDIBitmapInfoInternal *info, void *bitmap)
{
    Pixmap pixmap;
    spBool flag = SP_FALSE;
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    if (dibinstance->filename == NULL) return SP_FALSE;

    if (dibinstance->pixmap != None) {
	pixmap = dibinstance->pixmap;
    } else {
	pixmap = XCreatePixmap(SpTopLevelArch(sp_toplevel).display, 
			       RootWindowOfScreen(SpTopLevelArch(sp_toplevel).screen),
			       info->width, info->height,
			       DefaultDepthOfScreen(SpTopLevelArch(sp_toplevel).screen));
    }

    if (pixmap != None) {
	XPutImage(SpTopLevelArch(sp_toplevel).display,
		  pixmap,
		  SpTopLevelArch(sp_toplevel).fg_gc,
		  (XImage *)bitmap,
		  0, 0, 0, 0, info->width, info->height);
    
	flag = spSavePixmapToFileXm(pixmap, dibinstance->filename);
    
	if (dibinstance->pixmap != pixmap) {
	    XFreePixmap(SpTopLevelArch(sp_toplevel).display, pixmap);
	}
    }
    
    return flag;
}
